/* Header definitions and structures.
 * include/services.h
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact for contact information related to this software. */

/*************************************************************************/

#ifndef SERVICES_H
#define SERVICES_H

/*************************************************************************/

#include "sysconf.h"
#include "config.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <grp.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <setjmp.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#if HAVE_GETTIMEOFDAY
# include <sys/time.h>
#endif
#include <ctype.h>

#undef tolower
#undef toupper
#define tolower tolower_
#define toupper toupper_
extern int toupper(char), tolower(char);

/*************************************************************************/

#include "defs.h"
#include "language/numeric.h"

/*************************************************************************/

/* Configuration structure, all the options and values from the configuration 
 * file are put into this structure and used. */

struct conf {
   char rserver[HOSTMAX];
   int rport;
   char lpassword[PASSMAX];
   char lserver[HOSTMAX];
   char ldesc[DESCMAX];
   char *logfile;
   char *pidfile;
   char *motdfile;
   int servscan;
   int chanscan;
   char csnick[NICKMAX];
   char csident[IDENTMAX];
   char cshost[HOSTMAX];
   char csmodes[MODEMAX];
   char chandb[DBMAX];
   int csakicks;
   int csexpiry;
   char csakickreason[MSGMAX];
   int csinhabit;
   char nsnick[NICKMAX];
   char nsident[IDENTMAX];
   char nshost[HOSTMAX];
   char nsmodes[MODEMAX];
   char nickdb[DBMAX];
   int nsregdelay;
   int nsexpiry;
   int nsrelease;
   int nsstrict;
   int nslistoper;
   int nsbadpassword;
   int nsbadtimeout;
   char msnick[NICKMAX];
   char msident[IDENTMAX];
   char mshost[HOSTMAX];
   char msmodes[MODEMAX];
   char memodb[DBMAX];
   char cmemodb[DBMAX];
   int msmemos;
   int mssnddelay;
   int msmailactive;
   int msexpiry;
   char *msmailbinary;
   char *msmailfooter;
   char *msmailsubject;
   char osnick[NICKMAX];
   char osident[IDENTMAX];
   char oshost[HOSTMAX];
   char osmodes[MODEMAX];
   char admindb[DBMAX];
   char akilldb[DBMAX];
   char routedb[DBMAX];
   char osroot[NICKMAX];
   char osakillmsg[MSGMAX];
   char hsnick[NICKMAX];
   char hsident[IDENTMAX];
   char hshost[HOSTMAX];
   char hsmodes[MODEMAX];
   char gnnick[NICKMAX];
   char gnident[IDENTMAX];
   char gnhost[HOSTMAX];
   char gnmodes[MODEMAX];
   char dumpchannel[CHANMAX];
   int dumpchannelactive;
   char dumpchannelmodes[MODEMAX];
   char spcchannel[CHANMAX];
   int spcchannelactive;
   char spcchannelmodes[MODEMAX];
   int servicesactive;
   int csactive;
   int nsactive;
   int msactive;
   int osactive;   
   int hsactive;
   int sendpings;
   int sendpingstimeout;
   char logonmsg1[MSGMAX];
   char logonmsg2[MSGMAX];
   char logonmsg3[MSGMAX];
   char opermsg1[MSGMAX];
   char opermsg2[MSGMAX];
   char opermsg3[MSGMAX];
   char nsregistermemo[MSGMAX];
   char klineemail[HOSTMAX];
   char klineurl[HOSTMAX];
   int floodmsgs;
   int floodmsgstime;
   int floodsmsgs;
   char getroot[PASSMAX];
   int yunum;
   int yumax;
   int yuping;
   int yusendq;
   int yonum;
   int yomax;
   int yoping;
   int yosendq;
} conf;

struct save {
   int nick_timeout; 	
   int chan_timeout; 	
   int memo_timeout; 	
   int oper_timeout; 	
   int akill_timeout; 	
   int cycle;
} save;

struct exp {
   int nick_timeout; 	
   int chan_timeout; 	
   int cycle;
} exp;

struct hub {
   int tscur;
   int tsmin;
   int svinfo;
} hub;

/*************************************************************************/

/* Ping structure, send's pings to all servers at the defined interval, this
 * structure holds the time the last pings were sent, for proccessing the next
 * ping sequence. */

struct ping {
   time_t last_sent;
} ping;

/*************************************************************************/

/* DS command, define statistics.  This structure controls and stores values on various command counts,
 * and occurences within the actual runtime of DS. */

struct stats {
   int sadmin_akills;
   int csop_akills;
   int sroot_akills;
} stats;

/*************************************************************************/

/* Run structure, controls the current run state, and therefore acts upon this
 * defined run state. */

struct run {
   time_t started;
   long flags;
} run;

#define RUN_READONLY	0x00000001 /* Running in read only mode. */
#define RUN_STARTED	0x00000002 /* Running and started. */
#define RUN_LOOP		0x00000004 /* Running but quitting. */
#define RUN_ROUTING	0x00000008 /* Running and into a reroute. */
#define RUN_SYNCING	0x00000010 /* Synced to the network, denoted as 1 second or less ping. */
#define RUN_QUIT		0x00000020 /* Running is in quit mode. */
#define RUN_RESTART	0x00000040 /* Running is in restart mode. */
#define RUN_SAVING	0x00000080 /* Running is doing saving. */
#define RUN_CRASH		0x00000100 /* Running is in crash mode. */
#define RUN_LINKED	0x00000200 /* Running to show were linked. */
#define RUN_LINKED1	0x00000400 /* Running to show were linked. */
#define RUN_SYNCING1	0x00000800 /* Synced to the network, denoted as 1 second or less ping. */

/*************************************************************************/

/* Server info structure.  Each server structure is stored in one of 256
 * lists; the list is determined by the first character of the server.
 * Servers are stored in alphabetical order within lists. */

typedef struct serverinfo_ ServerInfo;
struct serverinfo_ {
   ServerInfo *next, *prev;
   char server[SERVERMAX];
   int usrcnt;
   char desc[DESCMAX];
   int hops;
   long flags;
   int ping;
};

#define SI_CONNECT	0x00000001  /* Server is connected to the network. */
#define SI_JUPE		0x00000002  /* Server is connected, but is jupitered. */
#define SI_NOOP		0x00000004  /* Server is set on noop, opering is disallowed. */

/*************************************************************************/

/* Nickname info structure.  Each nick structure is stored in one of 256
 * lists; the list is determined by the first character of the nick.  Nicks
 * are stored in alphabetical order within lists. */

typedef struct nickinfo_ NickInfo;
struct nickinfo_ {
   NickInfo *next, *prev;
   char nick[NICKMAX];
   char pass[PASSMAX];
   char *last_usermask;
   char *last_realname;
   time_t time_registered;
   time_t last_seen;
   long accesscount;	
   long channelcount;
   long linkcount;
   char **access;	
   long flags;		
   time_t id_timestamp;
   unsigned short memomax; /* reserved.*/
   unsigned short channelmax; /* reserved. */
   char master_nick[NICKMAX];
   char mark_nick[NICKMAX];
   char *url;
   char *email;
};

/*************************************************************************/

#define NI_KILLPROTECT		0x00000001 /* Kill others who take this nick. */
#define NI_SECURE			0x00000002 /* Nick is not recognized unless identified. */
#define NI_VERBOTEN		0x00000004 /* Nick may not be registered or used. */
#define NI_ENCRYPTEDPW		0x00000008 /* Nick password is encrypted. */
#define NI_MEMO_SIGNON		0x00000010 /* Notify of memos at signon and un-away. */
#define NI_MEMO_RECEIVE		0x00000020 /* Notify of new memos when sent. */
#define NI_PRIVATE		0x00000040 /* Nick is not shown in list, to users. */
#define NI_BOT			0x00000080 /* Nick is a registered bot. */
#define NI_IRCOP			0x00000100 /* Nick is an IRC Operator. */
#define NI_MARK			0x00000200 /* Nick has been marked. */
#define NI_EXPIRY			0x00000400 /* Nick expires or does not expire. */
#define NI_NOMEMOS		0x00000800 /* Nick does not recieve memos. */
#define NI_FROZEN			0x00001000 /* Nick is frozen. */
#define NI_NOOPS			0x00002000 /* Nick does not want on op lists. */
#define NI_LANG_ENGLISH 	0x00004000 /* Language selected is english. */
#define NI_LANG_FRENCH		0x00008000 /* Language selected is french. */
#define NI_LANG_SPANISH		0x00010000 /* Language selected is spanish. */
#define NI_LANG_ITALIAN		0x00020000 /* Language selected is italian. */
#define NI_LANG_GERMAN		0x00040000 /* Language selected is german. */
#define NI_LANG_NORWEGIAN	0x00080000 /* Language selected is norwegian. */
#define NI_GUESTPROTECT		0x10000000 /* Others who take this nick will be changed. */
#define NI_KILL_HELD		0x20000000 /* Nick is being held after a kill */
#define NI_RECOGNIZED		0x40000000 /* Nick is from a known address. */
#define NI_IDENTIFIED		0x80000000 /* Nick is identified for. */

#define NI_RES0010		0x00100000
#define NI_RES0020		0x00200000
#define NI_RES0040		0x00400000
#define NI_RES0080		0x00800000
#define NI_RES0100		0x01000000
#define NI_RES0200		0x02000000
#define NI_RES0400		0x04000000
#define NI_RES0800		0x08000000

/*************************************************************************/

/* Channel info structures.  Stored similarly to the nicks, except that
 * the second character of the channel name, not the first, is used to
 * determine the list. */

/* Access levels for users on channels. */
typedef struct {
    short level;
#define in_use is_nick
    short in_use;       
    char *name;
} ChanAccess;

/* Two levels also serve as exclusive boundaries for valid access levels. */
#define ACCESS_FOUNDER          10000	/* Founder access. */
#define ACCESS_INVALID	       -10000	/* Disabled settings. */

/* AutoKick data for channels. */
typedef struct {
    short is_nick;
    short pad;
    char *name;
    char *reason;
} AutoKick;

typedef struct chaninfo_ ChannelInfo;
struct chaninfo_ {
   ChannelInfo *next, *prev;
   char name[CHANMAX];
   char founder[NICKMAX];		
   char founderpass[PASSMAX];
   char *desc;
   time_t time_registered;
   time_t last_used;
   long accesscount;
   ChanAccess *access;			
   long akickcount;
   AutoKick *akick;
   short mlock_on, mlock_off;		
   long mlock_limit;			
   char *mlock_key;			
   char *last_topic;		
   char last_topic_setter[NICKMAX];	
   time_t last_topic_time;		
   long flags;				
   short *levels;			
   char *url;
   char *email;
   char mark_nick[NICKMAX];
   char join_message[MSGMAX];
   char part_message[MSGMAX];
   struct channel_ *c;
};

#define CI_KEEPTOPIC            0x00000001  /* Keep the topic after the last user parts. */
#define CI_SECUREOPS            0x00000002  /* Disallow non access users to be opped. */
#define CI_PRIVATE              0x00000004  /* Channel is not shown in list, to users. */
#define CI_TOPICLOCK            0x00000008  /* Topic may only be changed using set topic. */
#define CI_RESTRICTED           0x00000010  /* Only channel operators are allowed in. */
#define CI_LEAVEOPS             0x00000020  /* Do not deop everyone. */
#define CI_SECURE               0x00000040  /* Disallow access unless identified. */
#define CI_VERBOTEN             0x00000080  /* Channel may not be registered or used. */
#define CI_ENCRYPTEDPW          0x00000100  /* Channel password is encrypted. */
#define CI_FROZEN               0x00000200  /* Channel is frozen, access is disallowed. */
#define CI_IRCOP                0x00000400  /* Only IRC Operators may join. */
#define CI_ADMIN                0x00000800  /* Only Administrators may join. */
#define CI_CHELP                0x00001000  /* Channel is a registered help channel. */
#define CI_MARK                 0x00002000  /* Channel is marked, as well as nick who marked it. */
#define CI_EXPIRY               0x00004000  /* Channel expires or does not expire. */
#define CI_OPGUARD		  0x00008000  /* If an AOp or SOp is deoped by a non AOp or SOp. */ 
#define CI_VERBOSE              0x00040000
#define CI_CLOSE			  0x00010000  /* Channel has been closed by administration. */
#define CI_PROTECTED		  0x00020000  /* This chanel is protected from INFO commands. */

#define CI_RES0004	        0x00040000
#define CI_RES0008		  0x00080000
#define CI_RES0010		  0x00100000
#define CI_RES0020		  0x00200000
#define CI_RES0040		  0x00400000
#define CI_RES0080		  0x00800000
#define CI_RES0100		  0x01000000
#define CI_RES0200		  0x02000000
#define CI_RES0400		  0x04000000
#define CI_RES0800		  0x08000000
#define CI_RES1000		  0x10000000
#define CI_RES2000		  0x20000000
#define CI_RES4000		  0x40000000
#define CI_RES8000		  0x80000000

/* Access levels for a channel. */
#define CA_INVITE               0  /* May use the invite command. */
#define CA_AKICK                1  /* May use the akick command. */
#define CA_SET                  2  /* May use selected set commands. */
#define CA_UNBAN                3  /* May use the unban command. */
#define CA_AUTOOP               4  /* User get automatic channel operator status after joining. */
#define CA_AUTODEOP             5  /* Maximum, not minimum. */
#define CA_AUTOVOICE            6  /* Autovoice on join. */
#define CA_OPDEOP               7  /* May use op/deop. */
#define CA_ACCESS_LIST          8  /* May use access list. */
#define CA_CLEAR                9  /* May use clear. */
#define CA_NOJOIN               10 /* Maximum. */
#define CA_ACCESS_CHANGE        11 /* May use access add and del. */
#define CA_SEND_MEMO		  12 /* May send channel memos. */
#define CA_READ_MEMO		  13 /* May read channel memos. */
#define CA_DEL_MEMO		  14 /* May delete channel memos. */
#ifdef V_UNREAL
# define CA_AUTOHALFOP           15 /* User get automatic channel half operator status after joining. */
# define CA_SIZE                 16 /* Null pointer to end of access settings. */
#else
# define CA_SIZE                 15 /* Null pointer to end of access settings. */
#endif

/*************************************************************************/

/* MemoServ data.  Only nicks and channels that actually have memos and news get 
 * records in MemoServ's and NewsServ's lists, which are stored the same way
 * NickServ's are. There are a three places where memos are stored, the InBox and
 * Trash and Mail folders, all directly accessible. */

typedef struct memo_ Memo;
struct memo_ {
   char sender[NICKMAX];
   long number;				
   time_t time;				
   char *text;
   long flags;
   short reserved_s;			
   long reserved[3];			
};

#define MF_UNREAD		0x00000001 /* This memo has not yet been read. */

#define MF_T_SYS		0x00000002 /* This memo is a system memo. */
#define MF_T_NORM		0x00000004 /* This memo is a normal user memo. */ 
#define MF_T_OPER		0x00000008 /* This memo is an operator memo. */ 
#define MF_T_ADMIN	0x00000010 /* This memo is an administrator memo. */
#define MF_T_REPLY	0x00000020 /* This memo is a replied memo. */
#define MF_T_FORWARD	0x00000040 /* This memo is a forwarded memo. */

#define MF_DELETED	0x00000080 /* This memo is marked as deleted. */

#define MF_L_INBOX	0x00000100 /* Memo's stored in the InBox folder. */
#define MF_L_TRASH	0x00000200 /* Memo's stored in the Trash folder. */
#define MF_L_MAIL		0x00000400 /* Memo's stored in the Mail folder. */

#define MF_T_MAIL		0x00000800 /* This memo is a mail memo. */
#define MF_T_CHANNEL	0x00001000 /* This memo is a channel memo. */

#define MF_RES2		0x00002000
#define MF_RES4		0x00004000

/* MemoServ memo structure list for nick's who have memos. */
typedef struct memolist_ MemoList;
struct memolist_ {
   MemoList *next, *prev;
   char nick[NICKMAX];
   long n_memos;
   long n_mail;
   long n_trash;
   Memo *memos;
   Memo *mail;
   Memo *trash;
//   long reserved[4];
};

/* MemoServ memo structure list for channel's which have memos. */
typedef struct cmemolist_ CMemoList;
struct cmemolist_ {
   CMemoList *next, *prev;
   char chan[CHANMAX];	
   long n_memos;
   long n_trash;
   Memo *memos;
   Memo *trash;
//   long reserved[4];
};

/*************************************************************************/

/* Online user and channel structural data information, all connecting users and
 * created channels are stored in the lists below. */

typedef struct user_ User;
typedef struct channel_ Channel;
struct user_ {
   User *next, *prev;
   char nick[NICKMAX];
   NickInfo *ni;			
   char *username;
   char *host;				
   char *realname;
   char *server;			
   time_t signon;
   time_t my_signon;			
   int32 mode;
   struct u_chanlist {
      struct u_chanlist *next, *prev;
      Channel *chan;
   } *chans; /* List of channels currently joined. */
   struct u_chaninfolist {
      struct u_chaninfolist *next, *prev;
      ChannelInfo *chan;
   } *founder_chans; /* List of channels identified for. */
   struct u_nickinfolist {
      struct u_nickinfolist *next, *prev;
      NickInfo *nick;
   } *identified_nicks; /* List of nicks identified for. */
   short invalid_pw_count;		
   time_t invalid_pw_time;		
   time_t lastmemosend;		
   time_t lastnickreg;			
   long flags;
};

/*************************************************************************/

#define U_AWAY    0x00001000
#define U_NOOP    0x00002000

#define UMODE_O   0x00000001
#define UMODE_I   0x00000002
#define UMODE_S   0x00000004
#define UMODE_W   0x00000008
#define UMODE_G   0x00000010
#define UMODE_H   0x00000020
#define UMODE_A   0x00000040
#ifdef ELF
# define UMODE_X  0x00000080
# define UMODE_Y  0x00000100
#endif
#define UMODE_F   0x00000200
#define UMODE_K   0x00000400
#define UMODE_SA  0x00000800

/*************************************************************************/

struct channel_ {
   Channel *next, *prev;
   char name[CHANMAX];
   ChannelInfo *ci;			
   time_t creation_time;		
   char *topic;
   char topic_setter[NICKMAX];		
   time_t topic_time;			
   int32 mode;				
   int32 limit;			
   char *key;				
   int32 bancount, bansize;
   char **bans;
   struct c_userlist {
      struct c_userlist *next, *prev;
      User *user;
#ifdef V_UNREAL
   } *users, *chanops, *halfops, *voices;
#else
   } *users, *chanops, *voices;
#endif
};

/*************************************************************************/

/* Channel status flags, modes and presense. */
#define CMODE_I		0x00000001
#define CMODE_M		0x00000002
#define CMODE_N		0x00000004
#define CMODE_P		0x00000008
#define CMODE_S		0x00000010
#define CMODE_T		0x00000020
#define CMODE_K		0x00000040		
#define CMODE_L		0x00000080
#define CMODE_R		0x00000100
#ifdef V_BAHAMUT
# define CMODE_C		0x00000200
#endif
#ifdef V_UNREAL
# define CMODE_H		0x00000400
#endif

/*************************************************************************/

/* Services administrators info structure.  Each services administrator 
 * structure is stored in one of 256 lists; the list is determined by the
 * first character of the nick.  Nicks are stored in alphabetical order 
 * within lists. */

typedef struct admininfo_ AdminInfo;
struct admininfo_ {
   AdminInfo *next, *prev;
   char nick[NICKMAX];
   short level;
   short aop;
   long flags;
};

typedef struct clonelist_ CloneList;
struct clonelist_ {
   CloneList *next, *prev;
   char host[HOSTMAX];
   int count;
   int notice;
};

#define AI_AOP_SERVER	1
#define AI_AOP_OS		2
#define AI_AOP_CS		3

#define AI_P_CEO	0x00000002 /* This nick is the CEO of the network. */
#define AI_P_COO	0x00000004 /* This nick is the COO of the network. */
#define AI_P_CTO	0x00000008 /* This nick is the CTO of the network. */
#define AI_P_CRO	0x00000010 /* This nick is the CRO of the network. */
#define AI_P_CAO	0x00000020 /* This nick is the CAO of the network. */
#define AI_P_EB	0x00000040 /* This nick is on EB of the network. */

/*************************************************************************/

/* Services routing info structure.  Each route  structure is stored in 
 * one of 256 lists; the list is determined by the first character of
 * the server.  Servers are stored in alphabetical order within lists. */

typedef struct routeinfo_ RouteInfo;
struct routeinfo_ {
   RouteInfo *next, *prev;
   char server[HOSTMAX];
   char pass[PASSMAX];
   int port;
};

/*************************************************************************/

/* Hash structures, define tables which verify acces for the selected
 * command. */

typedef struct hash_ Hash;
struct hash_ {
    char *cmd;
    short level;
    void (*func)(const char *source);
    short identify;
};

typedef struct shash_ SHash;
struct shash_ {
    char *cmd;
    short level;
    void (*ms_func)(NickInfo *ni, char *param);
    void (*ns_func)(NickInfo *ni, char *param);
    void (*os_func)(NickInfo *ni, char *param);
    void (*cs_func)(User *u, ChannelInfo *ci, char *param);
    short identify;
};


/* Does this command require top level identification. */
#define H_IDENTIFY	1
#define H_NIDENTIFY	0

/* Exclusive levels for command processing and execution. */
#define SA_NONE		-1
#define SA_OPER		0
#define SA_ADMIN		1
#define SA_SOP		2
#define SA_SR		3
#define SA_HSR		4

/*************************************************************************/

#include "extern.h"
#include "proto.h"

/*************************************************************************/

#endif

/*************************************************************************/
