/* ChanServ functions.
 * src/chanserv.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to
 * this software. */

/*************************************************************************/

#ifndef CHANSERV_C
#define CHANSERV_C

/*************************************************************************/

#include "services.h"
#include "timeout.h"

/*************************************************************************/

ChannelInfo *chanlists[256];

static int def_levels[][2] = {
    { CA_AUTOVOICE,          		 		    5 }, 	/* Process for VOp as well. */
#ifdef V_UNREAL
    { CA_AUTOHALFOP,         				   10 }, 	/* Process for HOp as well. */
#endif
    { CA_AUTOOP,             				   15 }, 	/* Process for AOp as well. */
    { CA_INVITE,             				   15 },		
    { CA_UNBAN,              				   15 },
    { CA_OPDEOP,             				   15 },
    { CA_ACCESS_LIST,        				   15 },
    { CA_AKICK,             				   20 }, 	/* Process for SOp as well. */
    { CA_ACCESS_CHANGE,      				   20 },
    { CA_AUTODEOP,          				   -1 },
    { CA_NOJOIN,            				   -1 },
    { CA_SET,   				   ACCESS_INVALID },
    { CA_CLEAR, 				   ACCESS_INVALID },
    { CA_READ_MEMO,					    5 },
    { CA_SEND_MEMO,        	    			   20 },		
    { CA_DEL_MEMO,        	    			   20 },		
    { -1 }
};

typedef struct {
   int what;
   char *name;
   char *desc;
   int settable;
} LevelInfo;
static LevelInfo levelinfo[] = {
    { CA_AUTOOP,        	"AUTOOP",     	"Automatic channel operator status (AOp).",		C_TRUE	},
    { CA_AUTOVOICE,     	"AUTOVOICE",  	"Automatic channel voice status (VOp).",			C_TRUE	},
#ifdef V_UNREAL
    { CA_AUTOHALFOP,     	"AUTOHALFOP",  	"Automatic channel half operator status (HOp).",	C_TRUE	},
#endif
    { CA_AUTODEOP,      	"AUTODEOP",   	"Channel operator status disallowed.",			C_TRUE	},
    { CA_NOJOIN,        	"NOJOIN",     	"Not allowed to join channel if restricted.",		C_TRUE 	},
    { CA_INVITE,        	"INVITE",     	"May invite themself into a channel.",			C_TRUE 	},
    { CA_AKICK,         	"AKICK",      	"May AutoKick users from a channel (SOp).",		C_TRUE	},
    { CA_SET,           	"SET",        	"May set various options for the channel.",		C_TRUE	},
    { CA_CLEAR,         	"CLEAR",      	"May MassDeop and MassKick users from a channel.",	C_TRUE	},
    { CA_UNBAN,         	"UNBAN",      	"May remove bans place by ChanServ.", 			C_TRUE	},
    { CA_OPDEOP,        	"OPDEOP",    	"May op and deop users in a channel.", 			C_TRUE	},
    { CA_ACCESS_LIST,   	"ACCLIST",   	"May list a channels access list.", 			C_TRUE	},
    { CA_ACCESS_CHANGE, 	"ACCCHANGE", 	"May modify a channels access lists.", 			C_TRUE	},
    /* These are defined here and above, but are NOT settable, they are only here for information purposes
     * only.  They are statically bound to their respective levels for access above. */
    { CA_SEND_MEMO,        	"SENDMEMO",     	"May send channel memos (Bound to SOp).",			C_FALSE	},
    { CA_READ_MEMO,        	"READMEMO",     	"May read and list channel msmos (Bound to VOp).",	C_FALSE	},
    { CA_DEL_MEMO,        	"DELMEMO",     	"May delete channel memos (Bound to SOp).",		C_FALSE	},
    { -1 }
};
static int levelinfo_maxwidth = 0;

/*************************************************************************/

#include "cs-help.c"

/*************************************************************************/

static void alpha_insert_chan 		FUNCTION((ChannelInfo *ci));
static ChannelInfo *makechan 			FUNCTION((const char *chan));
static int delchan 				FUNCTION((ChannelInfo *ci));
static void reset_levels 			FUNCTION((ChannelInfo *ci));
static int is_founder 				FUNCTION((User *user, ChannelInfo *ci));
static int is_identified			FUNCTION((User *user, ChannelInfo *ci));
static int is_nick_identified 		FUNCTION((User *user, NickInfo *ni));
static int is_nick_access 			FUNCTION((User *user, char *nick));
static int get_access 				FUNCTION((User *user, ChannelInfo *ci));
static char *get_why_access 			FUNCTION((User *user, ChannelInfo *ci));
static void cs_help 				FUNCTION((const char *source));
static void cs_register 			FUNCTION((const char *source));
static void cs_identify 			FUNCTION((const char *source));
static void cs_drop 				FUNCTION((const char *source));
static void cs_set 				FUNCTION((const char *source));
static void cs_set_founder 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_password 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_desc 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_url 				FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_email 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_joinmsg 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_partmsg 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_topic 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_mlock 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_keeptopic 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_topiclock 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_private 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_secureops 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_restricted 		FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_secure 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_opguard 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_ircop 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_admin 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_protected 			FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_set_verbose                FUNCTION((User *u, ChannelInfo *ci, char *param));
static void cs_access 				FUNCTION((const char *source));
static void cs_aop 				FUNCTION((const char *source));
static void cs_sop 				FUNCTION((const char *source));
static void cs_vop 				FUNCTION((const char *source));
#ifdef V_UNREAL
   static void cs_hop 				FUNCTION((const char *source));
   static void cs_halfop 			FUNCTION((const char *source));
   static void cs_dehalfop			FUNCTION((const char *source));
#endif
static void cs_akick 				FUNCTION((const char *source));
static void cs_info 				FUNCTION((const char *source));
static void cs_list 				FUNCTION((const char *source));
static void cs_invite 				FUNCTION((const char *source));
static void cs_levels 				FUNCTION((const char *source));
static void cs_op 				FUNCTION((const char *source));
static void cs_deop 				FUNCTION((const char *source));
static void cs_voice 				FUNCTION((const char *source));
static void cs_devoice 				FUNCTION((const char *source));
static void cs_unban 				FUNCTION((const char *source));
static void cs_mdeop 				FUNCTION((const char *source));
static void cs_mkick 				FUNCTION((const char *source));
static void cs_getpass 				FUNCTION((const char *source));
static void cs_close 				FUNCTION((const char *source));
static void cs_forbid 				FUNCTION((const char *source));
static void cs_status 				FUNCTION((const char *source));
static void cs_mark 				FUNCTION((const char *source));
static void cs_expiry 				FUNCTION((const char *source));
static void cs_setpass 				FUNCTION((const char *source));
static void cs_chelp 				FUNCTION((const char *source));
static void cs_why 				FUNCTION((const char *source));
static void cs_freeze 				FUNCTION((const char *source));
static void cs_getkey 				FUNCTION((const char *source));

/*************************************************************************/

/* get_chanserv_stats()
 * Return information on memory use.  Assumes pointers are valid. */

void get_chanserv_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    ChannelInfo *ci;

    for (i = 0; i < 256; i++) {
	for (ci = chanlists[i]; ci; ci = ci->next) {
	    count++;
	    mem += sizeof(*ci);
	    if (ci->desc)
		mem += strlen(ci->desc)+1;
	    mem += ci->accesscount * sizeof(ChanAccess);
	    mem += ci->akickcount * sizeof(AutoKick);
	    for (j = 0; j < ci->akickcount; j++) {
		if (ci->akick[j].name)
		    mem += strlen(ci->akick[j].name)+1;
		if (ci->akick[j].reason)
		    mem += strlen(ci->akick[j].reason)+1;
	    }
	    if (ci->mlock_key)
		mem += strlen(ci->mlock_key)+1;
	    if (ci->last_topic)
		mem += strlen(ci->last_topic)+1;
	    if (ci->levels)
		mem += sizeof(*ci->levels) * CA_SIZE;
	}
    }
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/

/* Main ChanServ routine. */

void chanserv(const char *source, char *buf)
{
    char *cmd, *s;
    char slvl[32];

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;
    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_ChanServ, source, "\1PING %s", s);
    } else { 
	 Hash *cs, cs_hash_table[] = {
	    { "HELP",		SA_NONE, 	cs_help,		H_NIDENTIFY	},
	    { "REGISTER",		SA_NONE, 	cs_register,	H_NIDENTIFY	},
	    { "INFO", 		SA_NONE, 	cs_info,		H_NIDENTIFY	},
	    { "DROP", 		SA_NONE, 	cs_drop,		H_NIDENTIFY	},
	    { "CANCEL", 		SA_NONE, 	cs_drop,		H_NIDENTIFY	},
	    { "SET",		SA_NONE, 	cs_set,		H_NIDENTIFY	},
	    { "ACCESS", 		SA_NONE, 	cs_access,		H_NIDENTIFY	},
	    { "ACC", 		SA_NONE, 	cs_access,		H_NIDENTIFY	},
	    { "AKICK", 		SA_NONE, 	cs_akick,		H_NIDENTIFY	},
	    { "INVITE", 		SA_NONE, 	cs_invite,		H_NIDENTIFY	},
	    { "UNBAN", 		SA_NONE,	cs_unban,		H_NIDENTIFY	},
	    { "-b", 		SA_NONE,	cs_unban,		H_NIDENTIFY	},
	    { "LIST", 		SA_NONE,	cs_list,		H_NIDENTIFY	},
	    { "OP",	 		SA_NONE,	cs_op,		H_NIDENTIFY	},
	    { "DEOP", 		SA_NONE, 	cs_deop,		H_NIDENTIFY	},
	    { "+o",	 		SA_NONE,	cs_op,		H_NIDENTIFY	},
	    { "-o", 		SA_NONE, 	cs_deop,		H_NIDENTIFY	},
	    { "VOICE",	 	SA_NONE,	cs_voice,		H_NIDENTIFY	},
	    { "DEVOICE", 		SA_NONE, 	cs_devoice,		H_NIDENTIFY	},
	    { "+v",	 		SA_NONE,	cs_voice,		H_NIDENTIFY	},
	    { "-v", 		SA_NONE, 	cs_devoice,		H_NIDENTIFY	},
	    { "WHY", 		SA_NONE, 	cs_why,		H_NIDENTIFY	},
	    { "LEVELS",	 	SA_NONE, 	cs_levels,		H_NIDENTIFY	},
	    { "AOP", 		SA_NONE, 	cs_aop,		H_NIDENTIFY	},
	    { "SOP", 		SA_NONE, 	cs_sop,		H_NIDENTIFY	},
	    { "VOP", 		SA_NONE, 	cs_vop,		H_NIDENTIFY	},
#ifdef V_UNREAL
	    { "HOP", 		SA_NONE, 	cs_hop,		H_NIDENTIFY	},
	    { "HALFOP",	 	SA_NONE,	cs_halfop,		H_NIDENTIFY	},
	    { "DEHALFOP",		SA_NONE, 	cs_dehalfop,	H_NIDENTIFY	},
	    { "+h",	 		SA_NONE,	cs_halfop,		H_NIDENTIFY	},
	    { "-h", 		SA_NONE, 	cs_dehalfop,	H_NIDENTIFY	},
#endif
	    { "MDEOP", 		SA_NONE, 	cs_mdeop,		H_NIDENTIFY	},
	    { "MKICK", 		SA_NONE, 	cs_mkick,		H_NIDENTIFY	},
	    { "IDENTIFY", 	SA_NONE, 	cs_identify,	H_NIDENTIFY	},
	    { "AUTH", 		SA_NONE, 	cs_identify,	H_NIDENTIFY	},
	    { "ID", 		SA_NONE, 	cs_identify,	H_NIDENTIFY	},
	    { "CHELP",	 	SA_ADMIN, 	cs_chelp,		H_NIDENTIFY	},
	    { "STATUS", 		SA_ADMIN, 	cs_status,		H_NIDENTIFY	},
	    { "GETPASS", 		SA_SOP, 	cs_getpass,		H_NIDENTIFY	},
	    { "GETKEY", 		SA_SOP, 	cs_getkey,		H_NIDENTIFY	},
	    { "SETPASS", 		SA_SOP, 	cs_setpass,		H_NIDENTIFY	},
	    { "FREEZE",		SA_SOP, 	cs_freeze,		H_NIDENTIFY	},
	    { "FORBID", 		SA_SOP, 	cs_forbid,		H_NIDENTIFY	},
	    { "MARK",	 	SA_SOP, 	cs_mark,		H_NIDENTIFY	},
	    { "CLOSE", 		SA_SOP, 	cs_close,		H_NIDENTIFY	},
	    { "EXPIRY", 		SA_SR, 	cs_expiry,		H_NIDENTIFY	},
	    { NULL }
	};
	if (cs = do_hash(source, cmd, cs_hash_table)) {
	   switch (get_admin_level(source)) {
	       case 0: strscpy(slvl, "IRC Operator", 32); 			break;
	       case 1: strscpy(slvl, "Services Administrator", 32); 	break;
	       case 2: strscpy(slvl, "Services Operator", 32); 		break;
	       case 3: strscpy(slvl, "Services Root", 32); 			break;
		 default: break;
	   }
	   if (cs->level == SA_ADMIN && !is_services_admin(source)) {
	       notice(s_ChanServ, source, "Your level is %s, the required level to use this command is Services Administrator.", slvl);
	   } else if (cs->level == SA_SOP && !is_services_op(source)) {
	       notice(s_ChanServ, source, "Your level is %s, the required level to use this command is Services Operator.", slvl);
	   } else if (cs->level >= SA_SR && !is_services_root(source)) {
	       notice(s_ChanServ, source, "Your level is %s, the required level to use this command is Services Root.", slvl);
	   } else {
		(*cs->func) (source);
	   }		
	} else {
	   notice(s_ChanServ, source, "The command \2%s\2 is unknown, please type \2/msg %s HELP\2 for help with commands.",
         	cmd, s_ChanServ);
      }
   }
}

/*************************************************************************/

/* Load/save data files. */

void load_cs_dbase(void)
{
    FILE *f = fopen(conf.chandb, "r");
    int ver, i, j;
    ChannelInfo *ci;
    NickInfo *ni;

   if (!f)
     return;

	for (i = 33; i < 256; i++) {
	    while (fgetc(f) == 1) {
		ci = smalloc(sizeof(ChannelInfo));
		if (1 != fread(ci, sizeof(ChannelInfo), 1, f))
		    fatal_perror("Read error on %s", conf.chandb);
		/* Can't guarantee the file is in a particular order...
		 * (Well, we can, but we don't have to depend on it.) */
		alpha_insert_chan(ci);
		ci->desc = read_string(f, conf.chandb);
		if (ci->url)
		    ci->url = read_string(f, conf.chandb);
		if (ci->email)
		    ci->email = read_string(f, conf.chandb);
		if (ci->mlock_key)
		    ci->mlock_key = read_string(f, conf.chandb);
		if (ci->last_topic)
		    ci->last_topic = read_string(f, conf.chandb);
		ci->c = NULL;

		if (ci->accesscount) {
		    ChanAccess *access;
		    access = smalloc(sizeof(ChanAccess) * ci->accesscount);
		    ci->access = access;
		    if (ci->accesscount != fread(access, sizeof(ChanAccess),
							ci->accesscount, f))
			fatal_perror("Read error on ", conf.chandb);
		    for (j = 0; j < ci->accesscount; j++, access++)
			access->name = read_string(f, conf.chandb);
		    j = 0; access = ci->access;
		    while (j < ci->accesscount) {
			if (!access->in_use) {
			    ci->accesscount--;
			    free(access->name);
			    if (j < ci->accesscount)
				bcopy(access+1, access,	sizeof(*access) *
							(ci->accesscount - j));
			} else {
			    j++;
			    access++;
			}
		    }
		    if (ci->accesscount) {
			ci->access = srealloc(ci->access,
					sizeof(ChanAccess) * ci->accesscount);
		    } else {
			free(ci->access);
			ci->access = NULL;
		    }
		} else {
		    ci->access = NULL;
		} 

		if (ci->akickcount) {
		    AutoKick *akick;
		    akick = smalloc(sizeof(AutoKick) * ci->akickcount);
		    ci->akick = akick;
		    if (ci->akickcount != fread(akick, sizeof(AutoKick), ci->akickcount, f))
			fatal_perror("Read error on %s", conf.chandb);
		    for (j = 0; j < ci->akickcount; j++, akick++) {
			akick->name = read_string(f, conf.chandb);
			if (akick->reason)
			    akick->reason = read_string(f, conf.chandb);
		    }
		    j = 0; akick = ci->akick;
		    while (j < ci->akickcount) {
			if (akick->is_nick < 0) {
			    ci->akickcount--;
			    free(akick->name);
			    if (akick->reason)
				free(akick->reason);
			    if (j < ci->akickcount)
				bcopy(akick+1, akick, sizeof(*akick) *
							(ci->akickcount - j));
			} else {
			    j++;
			    akick++;
			}
		    }
		    if (ci->akickcount) {
			ci->akick = srealloc(ci->akick,
					sizeof(AutoKick) * ci->akickcount);
		    } else {
			free(ci->akick);
			ci->akick = NULL;
		    }
		} else {
		    ci->akick = NULL;
		} 

		if (ci->levels) {
		    int n_entries;
		    ci->levels = NULL;
		    reset_levels(ci);
		    n_entries = fgetc(f)<<8 | fgetc(f);
		    if (n_entries < 0)
			fatal_perror("Read error on %s", conf.chandb);
		    if (n_entries <= CA_SIZE) {
			fread(ci->levels, sizeof(short), n_entries, f);
		    } else {
			fread(ci->levels, sizeof(short), CA_SIZE, f);
			fseek(f, sizeof(short) * (n_entries - CA_SIZE), SEEK_CUR);
		    }
		} else {
		    reset_levels(ci);
		}
	    }
	}
    fclose(f);
}

/*************************************************************************/

void save_cs_dbase(void)
{
    FILE *f = fopen(conf.chandb, "w");
    int i, j;
    ChannelInfo *ci;
    time_t save_start, save_end;

   if (!f)
      return;
   save_start = time(NULL);
   globops(conf.lserver, "Performing channel database save...");
    for (i = 33; i < 256; i++) {
	for (ci = chanlists[i]; ci; ci = ci->next) {
	    fputc(1, f);
	    if (1 != fwrite(ci, sizeof(ChannelInfo), 1, f))
		fatal_perror("Write error on %s", conf.chandb);
	    write_string(ci->desc ? ci->desc : "", f, conf.chandb);
	    if (ci->url)
		write_string(ci->url, f, conf.chandb);
	    if (ci->email)
		write_string(ci->email, f, conf.chandb);
	    if (ci->mlock_key)
		write_string(ci->mlock_key, f, conf.chandb);
	    if (ci->last_topic)
		write_string(ci->last_topic, f, conf.chandb);

	    if (ci->accesscount) {
		ChanAccess *access = ci->access;
		if (ci->accesscount !=
			fwrite(access, sizeof(ChanAccess), ci->accesscount, f))
		    fatal_perror("Write error on %s", conf.chandb);
		for (j = 0; j < ci->accesscount; j++, access++) {
		    if (access->in_use)
			write_string(access->name, f, conf.chandb);
		    else
			write_string("", f, conf.chandb);
		}
	    }

	    if (ci->akickcount) {
		AutoKick *akick = ci->akick;
		if (ci->akickcount !=
			fwrite(akick, sizeof(AutoKick), ci->akickcount, f))
		    fatal_perror("Write error on %s", conf.chandb);
		for (j = 0; j < ci->akickcount; j++, akick++) {
		    if (akick->is_nick >= 0) {
			write_string(akick->name, f, conf.chandb);
			if (akick->reason)
			    write_string(akick->reason, f, conf.chandb);
		    } else {
			write_string("", f, conf.chandb);
			if (akick->reason)
			    write_string("", f, conf.chandb);
		    }
		}
	    }

	    if (ci->levels) {
		fputc(CA_SIZE >> 8, f);
		fputc(CA_SIZE & 0xFF, f);
		fwrite(ci->levels, sizeof(short), CA_SIZE, f);
	    }

	} /* for (chanlists[i]) */

	fputc(0, f);

    } /* for (i) */

   fclose(f);
   save_end = time(NULL) - save_start;
   globops(conf.lserver, "Channel database save completed in %d seconds.", save_end);
}

/*************************************************************************/

/* Check the current modes on a channel; if they conflict with a mode lock,
 * fix them. */

void check_modes(const char *chan)
{
    Channel *c = findchan(chan);
    ChannelInfo *ci;
    char newmodes[32], *newkey = NULL;
    long newlimit = 0;
    char *end = newmodes;
    int modes;
    int set_limit = 0, set_key = 0;

    if (!c || !(ci = c->ci))
	return;

    *end++ = '+';
    modes = ~c->mode & ci->mlock_on;
#ifdef V_BAHAMUT
    if (modes & CMODE_C) {
	*end++ = 'c';
	c->mode |= CMODE_C;
    }
#endif
#ifdef V_UNREAL
    if (modes & CMODE_H) {
	*end++ = 'h';
	c->mode |= CMODE_H;
    }
#endif
    if (modes & CMODE_I) {
	*end++ = 'i';
	c->mode |= CMODE_I;
    }
    if (modes & CMODE_R) {
	*end++ = 'r';
	c->mode |= CMODE_R;
    }
    if (modes & CMODE_M) {
	*end++ = 'm';
	c->mode |= CMODE_M;
    }
    if (modes & CMODE_N) {
	*end++ = 'n';
	c->mode |= CMODE_N;
    }
    if (modes & CMODE_P) {
	*end++ = 'p';
	c->mode |= CMODE_P;
    }
    if (modes & CMODE_S) {
	*end++ = 's';
	c->mode |= CMODE_S;
    }
    if (modes & CMODE_T) {
	*end++ = 't';
	c->mode |= CMODE_T;
    }

    if (ci->mlock_limit && ci->mlock_limit != c->limit) {
	*end++ = 'l';
	newlimit = ci->mlock_limit;
	c->limit = newlimit;
	set_limit = 1;
    }
    if (!c->key && ci->mlock_key) {
	*end++ = 'k';
	newkey = ci->mlock_key;
	c->key = sstrdup(newkey);
	set_key = 1;
    } else if (c->key && ci->mlock_key && strcmp(c->key, ci->mlock_key) != 0) {
	char *av[3];
	send_cmd(s_ChanServ, "MODE %s -k %s", c->name, c->key);
	av[0] = sstrdup(c->name);
	av[1] = sstrdup("-k");
	av[2] = sstrdup(c->key);
	do_cmode(s_ChanServ, 3, av);
	free(av[0]);
	free(av[1]);
	free(av[2]);
	free(c->key);
	*end++ = 'k';
	newkey = ci->mlock_key;
	c->key = sstrdup(newkey);
	set_key = 1;
    }

    if (end[-1] == '+')
	end--;

    *end++ = '-';
    modes = c->mode & ci->mlock_off;
#ifdef V_BAHAMUT
    if (modes & CMODE_C) {
	*end++ = 'c';
	c->mode &= ~CMODE_C;
    }
#endif
#ifdef V_UNREAL
    if (modes & CMODE_H) {
	*end++ = 'h';
	c->mode &= ~CMODE_H;
    }
#endif
    if (modes & CMODE_I) {
	*end++ = 'i';
	c->mode &= ~CMODE_I;
    }
    if (modes & CMODE_R) {
	*end++ = 'r';
	c->mode &= ~CMODE_R;
    }
    if (modes & CMODE_M) {
	*end++ = 'm';
	c->mode &= ~CMODE_M;
    }
    if (modes & CMODE_N) {
	*end++ = 'n';
	c->mode &= ~CMODE_N;
    }
    if (modes & CMODE_P) {
	*end++ = 'p';
	c->mode &= ~CMODE_P;
    }
    if (modes & CMODE_S) {
	*end++ = 's';
	c->mode &= ~CMODE_S;
    }
    if (modes & CMODE_T) {
	*end++ = 't';
	c->mode &= ~CMODE_T;
    }

    if (c->key && (ci->mlock_off & CMODE_K)) {
	*end++ = 'k';
	newkey = sstrdup(c->key);
	free(c->key);
	c->key = NULL;
	set_key = 1;
    }
    if (c->limit && (ci->mlock_off & CMODE_L)) {
	*end++ = 'l';
	c->limit = 0;
    }

    if (end[-1] == '-')
	end--;

    if (end == newmodes)
	return;
    *end = 0;
    if (set_limit) {
	if (set_key)
	    send_cmd(s_ChanServ, "MODE %s %s %d :%s", c->name,
				newmodes, newlimit, newkey ? newkey : "");
	else
	    send_cmd(s_ChanServ, "MODE %s %s %d", c->name, newmodes, newlimit);
    } else if (set_key)
	send_cmd(s_ChanServ, "MODE %s %s :%s", c->name,
				newmodes, newkey ? newkey : "");
    else
	send_cmd(s_ChanServ, "MODE %s %s", c->name, newmodes);

    if (newkey && !c->key)
	free(newkey);
}

/*************************************************************************/

/* Check whether a user is allowed to be opped on a channel; if they
 * aren't, deop them.  If serverop is 1, the +o was done by a server.
 * Return 1 if the user is allowed to be opped, 0 otherwise. */

int check_valid_op(User *user, const char *chan, int serverop)
{
    ChannelInfo *ci = cs_findchan(chan);

    if (!ci || ci->flags & CI_LEAVEOPS)
	return 1;

    if (is_services_op(user->nick))
	return 1;

    if ((ci->flags & CI_VERBOTEN) || (ci->flags & CI_CLOSE)) {
	/* check_kick() will get them out; we needn't explain. */
	send_cmd(s_ChanServ, "MODE %s -o %s", chan, user->nick);
	return 0;
    }

    if (ci->flags & CI_FROZEN) {
        notice(s_ChanServ, user->nick, "This channel is frozen and you may not acquire channel operator or voice status until the freeze has been disabled.");
	send_cmd(s_ChanServ, "MODE %s -o %s", chan, user->nick);
         return 0;
    }

    if (serverop && !check_access(user, ci, CA_AUTOOP)) {
	notice(s_ChanServ, user->nick,
		"This channel has been registered with %s, access priviledges apply.", s_ChanServ);
	send_cmd(s_ChanServ, "MODE %s -o %s", chan, user->nick);
	return 0;
    }

    if (check_access(user, ci, CA_AUTODEOP)) {
	send_cmd(s_ChanServ, "MODE %s -o %s", chan, user->nick);
	return 0;
    }

    return 1;
}

/*************************************************************************/

int check_should_srop(User *user, const char *chan)
{
    ChannelInfo *ci = cs_findchan(chan);
    AdminInfo *ai;
    NickInfo *ni;

    if (!ci)
	return 0;

    if (is_services_root(user->nick)) {
	ai = findadmin(user->nick);
	switch (ai->aop) {
		case AI_AOP_SERVER:
			send_cmd(NULL, "MODE %s +o %s", chan, user->nick);
			break;
		case AI_AOP_OS:
			send_cmd(s_OperServ, "MODE %s +o %s", chan, user->nick);
			break;
		case AI_AOP_CS:
			send_cmd(s_ChanServ, "MODE %s +o %s", chan, user->nick);
			break;
		default:
			send_cmd(s_ChanServ, "MODE %s +o %s", chan, user->nick);
			break;
	}	
	ci->last_used = time(NULL);
	return C_TRUE;
    }
    return C_FALSE;
}

/*************************************************************************/

#ifdef V_UNREAL
/* Check whether a user should be half opped on a channel, and if so, do it.
 * Return 1 if the user was half opped, 0 otherwise.  (Updates the channel's
 * last used time if the user was half opped.) */

int check_should_halfop(User *user, const char *chan)
{
    ChannelInfo *ci = cs_findchan(chan);
    NickInfo *ni;

    if (!ci || (ci->flags & CI_VERBOTEN) || (ci->flags & CI_CLOSE) || !ci->levels[CA_AUTOHALFOP] || *chan == '+')
	return 0;

    if (ci->flags & (CI_FROZEN | CI_CLOSE)) {
         return 0;
    }

    if (ci->flags & CI_SECURE) {
	if (!(ni = findnick(user->nick)) || !(ni->flags & NI_IDENTIFIED))
	    return 0;
    }

    if (check_access(user, ci, CA_AUTOHALFOP)) {
	send_cmd(s_ChanServ, "MODE %s +h %s", chan, user->nick);
	return 1;
    }

    return 0;
}
#endif

/*************************************************************************/

/* Check whether a user should be opped on a channel, and if so, do it.
 * Return 1 if the user was voiced, 0 otherwise. */

int check_should_op(User *user, const char *chan)
{
    ChannelInfo *ci = cs_findchan(chan);
    NickInfo *ni;

    if (!ci || (ci->flags & CI_VERBOTEN) || (ci->flags & CI_CLOSE) || !ci->levels[CA_AUTOOP] || *chan == '+')
	return 0;

    if (ci->flags & (CI_FROZEN | CI_CLOSE)) {
         return 0;
    }

    if (ci->flags & CI_SECURE) {
	if (!(ni = findnick(user->nick)) || !(ni->flags & NI_IDENTIFIED))
	    return 0;
    }

    if (check_access(user, ci, CA_AUTOOP)) {
	send_cmd(s_ChanServ, "MODE %s +o %s", chan, user->nick);
	ci->last_used = time(NULL);
	return 1;
    }

    return 0;
}

/*************************************************************************/

/* Check whether a user should be voiced on a channel, and if so, do it.
 * Return 1 if the user was voiced, 0 otherwise. */

int check_should_voice(User *user, const char *chan)
{
    ChannelInfo *ci = cs_findchan(chan);
    NickInfo *ni;

    if (!ci || (ci->flags & CI_VERBOTEN) || (ci->flags & CI_CLOSE) || !ci->levels[CA_AUTOVOICE] || *chan == '+')
	return 0;

    if (ci->flags & (CI_FROZEN | CI_CLOSE)) {
         return 0;
    }

    if (ci->flags & CI_SECURE) {
	if (!(ni = findnick(user->nick)) || !(ni->flags & NI_IDENTIFIED))
	    return 0;
    }

    if (check_access(user, ci, CA_AUTOVOICE)) {
	send_cmd(s_ChanServ, "MODE %s +v %s", chan, user->nick);
	return 1;
    }

    return 0;
}

/*************************************************************************/

int is_chan_aop(const char *nick, ChannelInfo *ci)
{
   User *u = finduser(nick);

   if (check_access(u, ci, CA_AUTOOP))
	return 1;
   return 0;
}

/*************************************************************************/

#ifdef V_UNREAL
int is_chan_hop(const char *nick, ChannelInfo *ci)
{
   User *u = finduser(nick);

   if (check_access(u, ci, CA_AUTOHALFOP))
	return 1;
   return 0;
}
#endif

/*************************************************************************/

int is_chan_vop(const char *nick, ChannelInfo *ci)
{
   User *u = finduser(nick);

   if (check_access(u, ci, CA_AUTOVOICE))
	return 1;
   return 0;
}

/*************************************************************************/

int is_chan_sop(const char *nick, ChannelInfo *ci)
{
   User *u = finduser(nick);

   if (check_access(u, ci, CA_AKICK))
	return 1;
   return 0;
}

/*************************************************************************/

/* Tiny helper routine to get ChanServ out of a channel after it went in. */

static void timeout_leave(Timeout *to)
{
    char *chan = to->data;
    send_cmd(s_ChanServ, "PART %s", chan);
    send_cmd(s_ChanServ, "MODE %s -isb *!*@*.*", chan);
    free(to->data);
}

/*************************************************************************/

/* check_kick()
 * Check whether a user is permitted to be on a channel.  If so, return 0;
 * else, kickban the user with an appropriate message (could be either
 * AKICK or restricted access) and return 1.  Note that this is called
 * _before_ the user is added to internal channel lists (so do_kick() is
 * not called). */

int check_kick(User *user, const char *chan)
{
    ChannelInfo *ci = cs_findchan(chan);
    Channel *c;
    AutoKick *akick;
    int i;
    NickInfo *ni;
    char *av[3], *mask, *reason;
    Timeout *t;
    int stay;

    if (!ci)
	return 0;

    if ((ci->flags & CI_IRCOP) && !is_oper(user->nick)) {
	mask = create_mask(user);
	reason = "Sorry, this channel may only be used by IRC Operators.";
	goto kick;
    }
    if ((ci->flags & CI_ADMIN) && !is_sadmin(user->nick)) {
	mask = create_mask(user);
	reason = "Sorry, this channel may only be used by Services Administrators.";
	goto kick;
    }
    if (ci->mlock_on & CMODE_I && !check_access(user, ci, CA_INVITE)) {
       if (!(c = findchan(chan))) {
		mask = create_mask(user);
		reason = "Sorry, this channel is invite only.";
		goto kick;
       }
    }
    if (is_services_op(user->nick))
       return 0;
    if (ci->flags & CI_CLOSE) {
	mask = create_mask(user);
	reason = "Sorry, this channel has been closed by administration.";
	goto kick;
    }
    if (ci->flags & CI_VERBOTEN) {
	mask = create_mask(user);
	reason = "Sorry, this channel is forbidden and may not be used.";
	goto kick;
    }
    ni = user->ni;
    if (ni && !(ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED)))
	ni = NULL;
    for (akick = ci->akick, i = 0; i < ci->akickcount; akick++, i++) {
	if ((akick->is_nick > 0 && ni && stricmp(akick->name, user->nick) == 0) || (!akick->is_nick && match_usermask(akick->name, user))) {
	    mask = akick->is_nick ? create_mask(user) : sstrdup(akick->name);
	    reason = akick->reason ? akick->reason : conf.csakickreason;
	    goto kick;
	}
    }
    if (!(ci->flags & CI_RESTRICTED))
	return 0;
    if (check_access(user, ci, CA_NOJOIN)) {
	mask = create_mask(user);
	reason = "Sorry, you are not permitted to be on this channel.";
	goto kick;
    }
    return 0;

kick:
    /* Remember that the user has not been added to our channel user list
     * yet, so we check whether the channel does not exist */
    stay = !findchan(chan);
    av[0] = sstrdup(chan);
    if (stay) {
	send_cmd(s_ChanServ, "JOIN %s", chan);
	send_cmd(NULL, "MODE %s +o %s  0", chan, s_ChanServ);
	strcpy(av[0], chan);
	av[1] = sstrdup("+o");
	av[2] = sstrdup(s_ChanServ);
	do_cmode(conf.lserver, 3, av);
	free(av[1]);
	free(av[2]);
	strcpy(av[0], chan);
    }
    strcpy(av[0], chan);
    av[1] = sstrdup("+b");
    av[2] = mask;
    send_cmd(s_ChanServ, "MODE %s +b %s  %lu", chan, av[2], time(NULL));
    do_cmode(s_ChanServ, 3, av);
    free(av[0]);
    free(av[1]);
    free(av[2]);
    send_cmd(s_ChanServ, "KICK %s %s :%s", chan, user->nick, reason);
    if (stay) {
	t = add_timeout(conf.csinhabit, timeout_leave, 0);
	t->data = sstrdup(chan);
    } else {
	send_cmd(s_ChanServ, "PART %s", chan);
    }
    return 1;
}

/*************************************************************************/

/* Record the current channel topic in the ChannelInfo structure. */

void record_topic(const char *chan)
{
    Channel *c;
    ChannelInfo *ci;

    c = findchan(chan);
    if (!c || !(ci = c->ci))
	return;
    if (ci->last_topic)
	free(ci->last_topic);
    if (c->topic)
	ci->last_topic = sstrdup(c->topic);
    else
	ci->last_topic = NULL;
    strscpy(ci->last_topic_setter, c->topic_setter, NICKMAX);
    ci->last_topic_time = c->topic_time;
}

/*************************************************************************/

/* Restore the topic in a channel when it's created, if we should. */

void restore_topic(const char *chan)
{
    Channel *c = findchan(chan);
    ChannelInfo *ci;

    if (!c || !(ci = c->ci) || !(ci->flags & CI_KEEPTOPIC))
	return;
    if (c->topic)
	free(c->topic);
    if (ci->last_topic) {
	c->topic = sstrdup(ci->last_topic);
	strscpy(c->topic_setter, ci->last_topic_setter, NICKMAX);
	c->topic_time = ci->last_topic_time;
    } else {
	c->topic = NULL;
	strscpy(c->topic_setter, s_ChanServ, NICKMAX);
    }
    send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s", chan,
		c->topic_setter, c->topic_time, c->topic ? c->topic : "");
}

/*************************************************************************/

/* See if the topic is locked on the given channel, and return 1 (and fix
 * the topic) if so. */

int check_topiclock(const char *chan)
{
    Channel *c = findchan(chan);
    ChannelInfo *ci;

    if (!c || !(ci = c->ci) || !(ci->flags & CI_TOPICLOCK))
	return 0;
    if (c->topic)
	free(c->topic);
    if (ci->last_topic)
	c->topic = sstrdup(ci->last_topic);
    else
	c->topic = NULL;
    strscpy(c->topic_setter, ci->last_topic_setter, NICKMAX);
    c->topic_time = ci->last_topic_time;
    send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s", chan,
		c->topic_setter, c->topic_time, c->topic ? c->topic : "");
    return 1;
}

/*************************************************************************/

/* expire_chans()
 * Remove all channels which have expired. */

void expire_chans()
{
    ChannelInfo *ci, *next;
    int i;
    const time_t expire_time = (conf.csexpiry * 24 * 60 * 60);
    time_t now = time(NULL);
    time_t expire_start, expire_end;
    int total_chans = 0, expired_chans = 0, cleaned_chans = 0;

   expire_start = time(NULL);
   globops(conf.lserver, "Performing channel database expire...");
   if (conf.csexpiry > 0) {
    for (i = 0; i < 256; i++) {
	for (ci = chanlists[i]; ci; ci = next) {
	    next = ci->next;
		total_chans++;
	    /* Expire channel's who have expiry enabled and are not forbidden, forbidden channels, see below. */
	    if (now - ci->last_used >= expire_time) {
		if (!(ci->flags & CI_VERBOTEN) && (ci->flags & CI_EXPIRY)) {
			log("Expiring channel %s", ci->name);
			clear_cmemos(ci->name);
			delchan(ci);
			expired_chans++;
		}
	    /* Since forbidden channels are infact registered channels, just restricted they can receieve memos, and 
	     * every cycle, remove the memos sent to this channel. */
		if ((ci->flags & CI_VERBOTEN)) {
			clear_cmemos(ci->name);
			cleaned_chans++;
		}
	    }
	}
    }
   }
   expire_end = time(NULL) - expire_start;
   globops(conf.lserver, "Channel database expire completed E(%d/%d) C(%d/%d) in %d seconds.", 
     expired_chans, total_chans, cleaned_chans, total_chans, expire_end);
}

/*************************************************************************/

void cs_remove_nick(const char *nick)
{
    int i, j;
    ChannelInfo *ci;
    ChanAccess *ca;

    for (i = 0; i < 256; i++) {
	for (ci = chanlists[i]; ci; ci = ci->next) {
          /* Is this nick the founder?  If so delete the channel. */
          if (stricmp(ci->founder, nick) == 0)
             delchan(ci);
          /* Not the founder, but on the AOp or SOp list?  Delete them anyways. */
	    for (ca = ci->access, j = ci->accesscount; j > 0; ca++, j--) {
		if (ca->in_use && stricmp(ca->name, nick) == 0)
		    ca->in_use = 0;
	    }
	}
    }
}

/*************************************************************************/

/* Return the ChannelInfo structure for the given channel, or NULL if the
 * channel isn't registered. */

ChannelInfo *cs_findchan(const char *chan)
{
    ChannelInfo *ci;

    for (ci = chanlists[tolower(chan[1])]; ci; ci = ci->next) {
	if (stricmp(ci->name, chan) == 0)
	    return ci;
    }
    return NULL;
}

/*************************************************************************/

/* alpha_insert_chan()
 * Insert a channel alphabetically into the database. */

static void alpha_insert_chan(ChannelInfo *ci)
{
    ChannelInfo *ptr, *prev;
    char *chan = ci->name;

    for (prev = NULL, ptr = chanlists[tolower(chan[1])]; ptr != NULL && stricmp(ptr->name, chan) < 0; prev = ptr, ptr = ptr->next)
	;
    ci->prev = prev;
    ci->next = ptr;
    if (!prev)
	chanlists[tolower(chan[1])] = ci;
    else
	prev->next = ci;
    if (ptr)
	ptr->prev = ci;
}

/*************************************************************************/

/* makechan()
 * Add a channel to the database.  Returns a pointer to the new ChannelInfo
 * structure if the channel was successfully registered, NULL otherwise.
 * Assumes channel does not already exist. */

static ChannelInfo *makechan(const char *chan)
{
    ChannelInfo *ci;

    ci = scalloc(sizeof(ChannelInfo), 1);
    strscpy(ci->name, chan, CHANMAX);
    ci->time_registered = time(NULL);
    reset_levels(ci);
    alpha_insert_chan(ci);
    return ci;
}

/*************************************************************************/

/* Remove a channel from the ChanServ database.  Return 1 on success, 0
 * otherwise. */

static int delchan(ChannelInfo *ci)
{
    int i;

    if (ci->c)
	ci->c->ci = NULL;
    if (ci->next)
	ci->next->prev = ci->prev;
    if (ci->prev)
	ci->prev->next = ci->next;
    else
	chanlists[tolower(ci->name[1])] = ci->next;
    if (ci->desc)
	free(ci->desc);
    if (ci->mlock_key)
	free(ci->mlock_key);
    if (ci->last_topic)
	free(ci->last_topic);
    for (i = 0; i < ci->accesscount; i++) {
	if (ci->access[i].name)
	    free(ci->access[i].name);
    }
    if (ci->access)
	free(ci->access);
    for (i = 0; i < ci->akickcount; ++i) {
	if (ci->akick[i].name)
	    free(ci->akick[i].name);
	if (ci->akick[i].reason)
	    free(ci->akick[i].reason);
    }
    if (ci->akick)
	free(ci->akick);
    if (ci->levels)
	free(ci->levels);
    free(ci);
    return 1;
}

/*************************************************************************/

/* Reset channel access level values to their default state. */

static void reset_levels(ChannelInfo *ci)
{
    int i;

    if (ci->levels)
	free(ci->levels);
    ci->levels = smalloc(CA_SIZE * sizeof(*ci->levels));
    for (i = 0; def_levels[i][0] >= 0; i++)
	ci->levels[def_levels[i][0]] = def_levels[i][1];
}

/*************************************************************************/

/* Return the level associated with the level keyword. */

int ret_level(int level)
{
   int i;

   for (i = 0; def_levels[i][0] >= 0; i++)
      if (def_levels[i][0] == level)
         return def_levels[i][1];
}

/*************************************************************************/

/* Does the given user have founder access to the channel? */

static int is_founder(User *user, ChannelInfo *ci)
{
    NickInfo *lni, *ni = user->ni;
    
    if (is_identified(user, ci))
	return 1;
    if (ni && ((ni->flags & NI_IDENTIFIED) || ((ni->flags & NI_RECOGNIZED) && !(ci->flags & CI_SECURE)))
	   && stricmp(user->nick, ci->founder) == 0)
	return 1;
    
    return 0;
}

/*************************************************************************/

/* Has the given user password-identified as founder for the channel? */

static int is_identified(User *user, ChannelInfo *ci)
{
    struct u_chaninfolist *c;
    NickInfo *lni, *ni = findnick(user->nick);
    struct u_nickinfolist *u;

    for (c = user->founder_chans; c; c = c->next) {
	if (c->chan == ci)
	    return 1;
    }
    for (u = user->identified_nicks; u; u = u->next) {
       if (ni && (ni->flags & NI_IDENTIFIED))
		if (stricmp(u->nick->nick, ci->founder) == 0)
			return 1;
    }
    return 0;
}

/*************************************************************************/

/* Has the given user password-identified as owner secondary of a nick? */

static int is_nick_identified(User *user, NickInfo *ni)
{
    struct u_nickinfolist *u;
    
    for (u = user->identified_nicks; u; u = u->next) {
	if (u->nick == ni)
	    return 1;
    }
    return 0;
}

/*************************************************************************/

/* Has the given user password-identified as owner secondary of a nick? */

static int is_nick_access(User *user, char *nick)
{
    struct u_nickinfolist *u;
    
    for (u = user->identified_nicks; u; u = u->next) {
	if (stricmp(u->nick->nick, nick) == 0)
	    return 1;
    }
    return 0;
}

/*************************************************************************/

/* Return the access level the given user has on the channel.  If the
 * channel doesn't exist, the user isn't on the access list, or the channel
 * is CS_SECURE and the user hasn't IDENTIFY'd with NickServ, return 0. */

static int get_access(User *user, ChannelInfo *ci)
{
    NickInfo *ni, *lni;
    ChanAccess *access;
    int i;
    struct u_nickinfolist *u;

    if (!ci)
	return 0;
    if (is_founder(user, ci))
	return ACCESS_FOUNDER;
    if ((ni = user->ni) && ((ni->flags & NI_IDENTIFIED) || ((ni->flags & NI_RECOGNIZED) && !(ci->flags & CI_SECURE))) ) {
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if ((access->in_use && stricmp(user->nick, access->name) == 0))
		return access->level;
	}
    }
    /* Identified nicks, checked here. */
    for (u = user->identified_nicks; u; u = u->next) {
       for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    ni = findnick(user->nick);
	    if (ni && (ni->flags & NI_IDENTIFIED) && !(ci->flags & CI_SECURE)) {
		  if (stricmp(u->nick->nick, access->name) == 0)
			return access->level;       
          }
       }
    }
    return (ci->flags & CI_SECUREOPS) ? -1 : 0;
}

/*************************************************************************/

/* get_why_access()
 * This is a special get_access, it returns the access level and the nick that
 * has the access level.  This is used to provide more accurate data regarding nicks
 * who have access.  Otherwise it acts exactly like get_access(). */

static char *get_why_access(User *user, ChannelInfo *ci)
{
    NickInfo *ni, *lni;
    ChanAccess *access;
    int i;
    struct u_nickinfolist *u;

    if (!ci) {
	return NULL;
    }
    if (is_founder(user, ci)) {
	return (user->nick);
   }
    if ((ni = user->ni) && ((ni->flags & NI_IDENTIFIED) || ((ni->flags & NI_RECOGNIZED) && !(ci->flags & CI_SECURE))) ) {
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if ((access->in_use && stricmp(user->nick, access->name) == 0)) {
			return access->name;
	    }
	}
    }
    /* Identified nicks, checked here. */
    for (u = user->identified_nicks; u; u = u->next) {
       for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    ni = findnick(user->nick);
	    if (ni && (ni->flags & NI_IDENTIFIED) && !(ci->flags & CI_SECURE)) {
		  if (stricmp(u->nick->nick, access->name) == 0) {
			return access->name;       
		  }
          }
       }
    }
    return NULL;
}

/*************************************************************************/

/* Return 1 if the user's access level on the given channel falls into the
 * given category, 0 otherwise.  Note that this may seem slightly confusing
 * in some cases: for example, check_access(..., CA_NOJOIN) returns true if
 * the user does _not_ have access to the channel (i.e. matches the NOJOIN
 * criterion). */

int check_access(User *user, ChannelInfo *ci, int what)
{
    int level = get_access(user, ci);

    if (level == ACCESS_FOUNDER)
	return (what==CA_AUTODEOP || what==CA_NOJOIN) ? 0 : 1;
    if (ci->levels[what] == ACCESS_INVALID)
	return 0;
    if (what == CA_AUTODEOP || what == CA_NOJOIN)
	return level <= ci->levels[what];
    else
	return level >= ci->levels[what];
}

/*************************************************************************/

static void cs_help(const char *source)
{
    char *cmd = strtok(NULL, "");
    char buf[256];
    int is_servadmin = is_services_admin(source);

    if (cmd && is_servadmin) {

	if (stricmp(cmd, "DROP") == 0) {
	    notice_list(s_ChanServ, source, oper_drop_help);
	    return;
	} else if (stricmp(cmd, "GETPASS") == 0) {
	    notice_list(s_ChanServ, source, getpass_help);
	    return;
	} else if (stricmp(cmd, "FORBID") == 0) {
	    notice_list(s_ChanServ, source, forbid_help);
	    return;
	} else if (stricmp(cmd, "STATUS") == 0) {
	    notice_list(s_ChanServ, source, status_help);
	    return;
	}

    }

    if (cmd && strnicmp(cmd, "levels", 6) == 0 && isspace(cmd[6])) {
	char *s = cmd+7;
	while (*s && isspace(*s))
	    s++;
	if (stricmp(s, "desc") == 0) {
	    int i;
	    notice(s_ChanServ, source,
		"The following feature/function names are understood.  Note");
	    notice(s_ChanServ, source,
		"that the levels for AUTODEOP and NOJOIN are maximum levels,");
	    notice(s_ChanServ, source,
		"while all others are minimum levels.");
	    if (!levelinfo_maxwidth) {
		for (i = 0; levelinfo[i].what >= 0; i++) {
		    int len = strlen(levelinfo[i].name);
		    if (len > levelinfo_maxwidth)
			levelinfo_maxwidth = len;
		}
	    }
	    for (i = 0; levelinfo[i].what >= 0; i++)
		notice(s_ChanServ, source, "    %-*s  %s", levelinfo_maxwidth,
			levelinfo[i].name, levelinfo[i].desc);
	    return;
	}
    }

    snprintf(buf, sizeof(buf), "%s%s", s_ChanServ, cmd ? " " : "");
    strscpy(buf+strlen(buf), cmd ? cmd : "", sizeof(buf)-strlen(buf));
    helpserv(s_ChanServ, source, buf);
    if (!cmd && is_servadmin)
	notice_list(s_ChanServ, source, cs_servadmin_help);
}

/*************************************************************************/

static void cs_register(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    char *desc = strtok(NULL, "");
    NickInfo *ni = findnick(source);
    Channel *c;
    ChannelInfo *ci;
    User *u = finduser(source);
    struct u_chaninfolist *uc;

    if (!desc) {
	notice(s_ChanServ, source, "Syntax: REGISTER <channel> <password> <description>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP REGISTER\2.",
		s_ChanServ);
    } else if (*chan == '&') {
	notice(s_ChanServ, source, "Local channels can not be registered.");
    } else if (chan[1] == 0) {
	notice(s_ChanServ, source, "\2%s\2 is not a valid channel name.", chan);
    } else if (!ni) {
	notice(s_ChanServ, source, "You must register your nick before registering a channel.");
	notice(s_ChanServ, source, "Please type \2/msg %s HELP\2 for information on registering a nick.",
		s_NickServ);
    } else if (!(ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED))) {
	notice(s_ChanServ, source, "You must identify for your nick with %s first.",
		s_NickServ);
	notice(s_ChanServ, source, "Please type \2/msg %s IDENTIFY <password>\2", s_NickServ);
    } else if (ci = cs_findchan(chan)) {

	if (ci->flags & CI_VERBOTEN) {
	    log("%s: Attempt to register FORBIDden channel %s by %s!%s@%s",
			s_ChanServ, chan, source, u->username, u->host);
	    notice(s_ChanServ, source, "Channel \2%s\2 may not be registered.", chan);
	} else {
	    notice(s_ChanServ, source, "Channel \2%s\2 already registered.", chan);
	}
    } else if (!u) {
	log("%s: Attempt to register channel %s from nonexistent nick %s", s_ChanServ, chan, source);
	notice(s_ChanServ, source, "Sorry, registration failed.");
    } else if (!is_chanop(source, chan) && !is_services_op(source)) {
 	notice(s_ChanServ, source, "You must be a channel operator to register the channel.");
    } else if (!(c = findchan(chan))) {
	log("%s: Channel %s not found for REGISTER", s_ChanServ, chan);
	notice(s_ChanServ, source, "Sorry, registration failed.");
    } else if (!(ci = makechan(chan))) {
	log("%s: makechan() failed for REGISTER %s", s_ChanServ, chan);
	notice(s_ChanServ, source, "Sorry, registration failed.");
    } else {
	c->ci = ci;
	ci->c = c;
	ci->flags = CI_KEEPTOPIC | CI_EXPIRY;
      /* Set +r for a registered channel. */
	ci->mlock_on = CMODE_N | CMODE_T | CMODE_R;
	ci->last_used = ci->time_registered;
	strscpy(ci->founder, source, NICKMAX);
	strscpy(ci->founderpass, pass, PASSMAX);
	ci->desc = sstrdup(desc);
	if (c->topic) {
	    ci->last_topic = sstrdup(c->topic);
	    strscpy(ci->last_topic_setter, c->topic_setter, NICKMAX);
	    ci->last_topic_time = c->topic_time;
	}
	log("%s: Channel %s registered by %s!%s@%s", s_ChanServ, chan,
		source, u->username, u->host);
      clear_cmemos(ci->name);
	notice(s_ChanServ, source, "Channel \2%s\2 is now registered under your nick \2%s\2.", chan, source);
      notice(s_ChanServ, source, "Your password is \2%s\2, write it down and remember it for later use.", pass);
      notice(s_ChanServ, source, "This channel will expire if not used for a period of \2%d\2 days.", conf.csexpiry);
	uc = smalloc(sizeof(*uc));
	uc->next = u->founder_chans;
	uc->prev = NULL;
	if (u->founder_chans)
	    u->founder_chans->prev = uc;
	u->founder_chans = uc;
	uc->chan = ci;
	/* Implement new mode lock */
	check_modes(ci->name);
    }
}

/*************************************************************************/

static void cs_identify(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    ChannelInfo *ci;
    struct u_chaninfolist *uc;
    User *u = finduser(source);
    int i = 0;

    if (!pass) {
	notice(s_ChanServ, source, "Syntax: IDENTIFY %s <password>", (chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP IDENTIFY\2.",
		s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u) {
	log("%s: IDENTIFY from nonexistent nick %s for %s", s_ChanServ, source, chan);
	notice(s_ChanServ, source, "Could not retrieve your online user information.");
    } else {
	int res;
      for (uc = u->founder_chans; uc; uc = uc->next) {
	   if (uc->chan == ci)
	      i = 1;
      }
 
	if ((strcmp(pass, ci->founderpass)) == 0 && !i) {
	    if (!is_identified(u, ci)) {
		uc = smalloc(sizeof(*uc));
		uc->next = u->founder_chans;
		uc->prev = NULL;
		if (u->founder_chans)
		    u->founder_chans->prev = uc;
		u->founder_chans = uc;
		uc->chan = ci;
		log("%s: %s!%s@%s identified for %s", s_ChanServ,
			source, u->username, u->host, chan);
	    }
	    notice(s_ChanServ, source,
		"Password accepted for the channel \2%s\2, you now have founder access.",
		chan);
          if (conf.dumpchannelactive)
	      privmsg(s_ChanServ, conf.dumpchannel, "*%s* -> IDENTIFY %s %s", source, chan, pass);
	} else if (i) {
		notice(s_ChanServ, source, "You have already identified for the channel \2%s\2.", chan);
	} else {
	    log("%s: Failed IDENTIFY for %s by %s!%s@%s",
			s_ChanServ, chan, source, u->username, u->host);
	    notice(s_ChanServ, source, "Password incorrect.");
	    bad_password(u);
	}

    }
}

/*************************************************************************/

static void cs_drop(const char *source)
{
    char *chan = strtok(NULL, " ");
    Channel *c;
    ChannelInfo *ci;
    NickInfo *ni;
    User *u = finduser(source);
    int is_servop = is_services_op(source);

    if (!chan) {
	notice(s_ChanServ, source, "Syntax: DROP <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP DROP\2.",
			s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!is_servop && (!u || !is_identified(u, ci))) {
	notice(s_ChanServ, source, "Password authentication is required for that command.");
	notice(s_ChanServ, source, "Please type \2/msg %s IDENTIFY %s <password>.",
		s_ChanServ, chan);
    } else if (is_servop && !u) {
	log("%s: DROP %s from nonexistent oper %s", s_ChanServ, chan, source);
	notice(s_ChanServ, source, "Could not retrieve your online user information.");
    } else {
	ni = findnick(ci->founder);
	if (!ni) {
	    log("%s: Can't find founder nickname record for dropping %s",
			s_ChanServ, chan);
	}
	delchan(ci);
	log("%s: Channel %s dropped by %s!%s@%s", s_ChanServ, chan,
			source, u->username, u->host);
      if (is_servop)
         globops(s_ChanServ, "\2%s\2 has forcibly dropped (drop) the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Channel %s has been dropped.", chan);
      if (conf.dumpchannelactive)
 		privmsg(s_ChanServ, conf.dumpchannel, "*%s* -> DROP %s", source, chan);
    }
}

/*************************************************************************/

/* cs_set()
 * Main SET routine.  Calls other routines as follows:
 * cs_set_command(User *command_sender, ChannelInfo *ci, char *param);
 * The parameter passed is the first space-delimited parameter after the
 * option name, except in the case of DESC and TOPIC, in which it is the
 * entire remainder of the line.  Additional parameters beyond the first
 * passed in the function call can be retrieved using strtok(NULL, toks). */

static void cs_set(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *cmd  = strtok(NULL, " ");
    char *param;
    ChannelInfo *ci;
    User *u;
    int access;

    if (cmd) {
	if (stricmp(cmd, "DESC") == 0 || stricmp(cmd, "TOPIC") == 0 || stricmp(cmd, "JOIN") == 0 || stricmp(cmd, "PART") == 0)
	    param = strtok(NULL, "");
	else
	    param = strtok(NULL, " ");
    } else {
	param = NULL;
    }

    if (!param) {
	notice(s_ChanServ, source, "Syntax: SET %s <option> <parameters>", (chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP SET\2.", s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!(u = finduser(source)) || (access = get_access(u, ci), !check_access(u, ci, CA_SET))) {
	notice(s_ChanServ, source, "Access denied.");
    } else if (stricmp(cmd, "FOUNDER") == 0) {
	if (access < ACCESS_FOUNDER) {
	    notice(s_ChanServ, source, "Password authentication required for that command.");
	    notice(s_ChanServ, source, "Please type \2/msg %s IDENTIFY %s <password>.", s_ChanServ, chan);
	} else {
	    cs_set_founder(u, ci, param);
	}
    } else if (stricmp(cmd, "PASSWORD") == 0) {
	if (access < ACCESS_FOUNDER) {
	    notice(s_ChanServ, source, "Password authentication required for that command.");
	    notice(s_ChanServ, source, "Please type \2/msg %s IDENTIFY %s <password>.", s_ChanServ, chan);
	} else {
	    cs_set_password(u, ci, param);
	}
    } else if (stricmp(cmd, "DESC") == 0) {
	cs_set_desc(u, ci, param);
    } else if (stricmp(cmd, "URL") == 0) {
	cs_set_url(u, ci, param);
    } else if (stricmp(cmd, "EMAIL") == 0) {
	cs_set_email(u, ci, param);
    } else if (stricmp(cmd, "PART") == 0) {
	cs_set_partmsg(u, ci, param);
    } else if (stricmp(cmd, "JOIN") == 0) {
	cs_set_joinmsg(u, ci, param);
    } else if (stricmp(cmd, "TOPIC") == 0) {
	cs_set_topic(u, ci, param);
    } else if (stricmp(cmd, "MLOCK") == 0) {
	cs_set_mlock(u, ci, param);
    } else if (stricmp(cmd, "KEEPTOPIC") == 0) {
	cs_set_keeptopic(u, ci, param);
    } else if (stricmp(cmd, "TOPICLOCK") == 0) {
	cs_set_topiclock(u, ci, param);
    } else if (stricmp(cmd, "PRIVATE") == 0) {
	cs_set_private(u, ci, param);
    } else if (stricmp(cmd, "SECUREOPS") == 0) {
	cs_set_secureops(u, ci, param);
    } else if (stricmp(cmd, "RESTRICTED") == 0) {
	cs_set_restricted(u, ci, param);
    } else if (stricmp(cmd, "OPGUARD") == 0) {
	cs_set_opguard(u, ci, param);
    } else if (stricmp(cmd, "SECURE") == 0) {
	cs_set_secure(u, ci, param);
    } else if (stricmp(cmd, "VERBOSE") == 0) {
	cs_set_verbose(u, ci, param);
    } else if (!is_oper(u->nick)) {
      notice(s_ChanServ, source, "Your level is user, the required level to use this command is IRC Operator.");
    } else if (stricmp(cmd, "IRCOP") == 0) {
	cs_set_ircop(u, ci, param);
    } else if (!is_sadmin(u->nick)) {
      notice(s_ChanServ, source, "Your level is IRC Operator, the required level to use this command is Services Administrator.");
    } else if (stricmp(cmd, "ADMIN") == 0) {
	cs_set_admin(u, ci, param);
    } else {
      notice(s_ChanServ, source, "The set command \2%s\2 is unknown, please type \2/msg %s HELP SET\2 for help with commands.", cmd, s_ChanServ);
    }
}

/*************************************************************************/

static void cs_set_founder(User *u, ChannelInfo *ci, char *param)
{
    NickInfo *ni = findnick(param);

    if (!ni) {
	notice(s_ChanServ, u->nick, "Nick \2%s\2 is not registered.", param);
    } else {
      strscpy(ci->founder, param, NICKMAX);
	notice(s_ChanServ, u->nick, "Founder of the channel \2%s\2 changed to \2%s\2.", ci->name, ci->founder);
    }
}

/*************************************************************************/

static void cs_set_password(User *u, ChannelInfo *ci, char *param)
{
    strscpy(ci->founderpass, param, PASSMAX);
    notice(s_ChanServ, u->nick, "Password for the channel \2%s\2 changed to \2%s\2.", ci->name, ci->founderpass);
}

/*************************************************************************/

static void cs_set_desc(User *u, ChannelInfo *ci, char *param)
{
    if (ci->desc)
	free(ci->desc);
    ci->desc = sstrdup(param);
    notice(s_ChanServ, u->nick, "Description of the channel \2%s\2 changed to \2%s\2.", ci->name, param);
}

/*************************************************************************/

static void cs_set_url(User *u, ChannelInfo *ci, char *param)
{
    if (ci->url)
	free(ci->url);
    ci->url = sstrdup(param);
    notice(s_ChanServ, u->nick, "URL for the channel \2%s\2 changed to \2%s\2.", ci->name, param);
}

/*************************************************************************/

static void cs_set_email(User *u, ChannelInfo *ci, char *param)
{
    if (ci->email)
	free(ci->email);
    ci->email = sstrdup(param);
    notice(s_ChanServ, u->nick, "E-mail address for the channel \2%s\2 changed to \2%s\2.", ci->name, param);
}

/*************************************************************************/

static void cs_set_joinmsg(User *u, ChannelInfo *ci, char *param)
{
    strscpy(ci->join_message, param, MSGMAX);
    notice(s_ChanServ, u->nick, "Join message for the channel \2%s\2 has been changed.", ci->name);
}

/*************************************************************************/

static void cs_set_partmsg(User *u, ChannelInfo *ci, char *param)
{
    strscpy(ci->part_message, param, MSGMAX);
    notice(s_ChanServ, u->nick, "Part message for the channel \2%s\2 has been changed.", ci->name);
}

/*************************************************************************/

static void cs_set_topic(User *u, ChannelInfo *ci, char *param)
{
    Channel *c = ci->c;

    if (!c) {
	log("%s: SET TOPIC for %s from %s: channel not found!", s_ChanServ, ci->name, u->nick);
	notice(s_ChanServ, u->nick, "Channel \2%s\2 does not exist.", ci->name);
	return;
    }
    if (ci->last_topic)
	free(ci->last_topic);
    if (*param)
	ci->last_topic = sstrdup(param);
    else
	ci->last_topic = NULL;
    if (c->topic) {
	free(c->topic);
	c->topic_time--;	/* to get around TS8 */
    } else
	c->topic_time = time(NULL);
    if (*param)
	c->topic = sstrdup(param);
    else
	c->topic = NULL;
    strscpy(ci->last_topic_setter, u->nick, NICKMAX);
    strscpy(c->topic_setter, u->nick, NICKMAX);
    ci->last_topic_time = c->topic_time;
    send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s", ci->name, u->nick, c->topic_time, param);
}

/*************************************************************************/

static void cs_set_mlock(User *u, ChannelInfo *ci, char *param)
{
    char *s, modebuf[32], *end, c;
    int add = -1;	/* 1 if adding, 0 if deleting, -1 if neither */
    int newlock_on = 0, newlock_off = 0, newlock_limit = 0;
    char *newlock_key = NULL;

    while (*param) {
	if (*param != '+' && *param != '-' && add < 0) {
	    param++;
	    continue;
	}
	switch (tolower((c = *param++))) {
	  case '+':
	    add = 1;
	    break;
	  case '-':
	    add = 0;
	    break;
#ifdef V_BAHAMUT
	  case 'c':
	    if (add) {
		newlock_on |= CMODE_C;
		newlock_off &= ~CMODE_C;
	    } else {
		newlock_off |= CMODE_C;
		newlock_on &= ~CMODE_C;
	    }
	    break;
#endif
#ifdef V_UNREAL
	  case 'h':
	    if (add) {
		newlock_on |= CMODE_H;
		newlock_off &= ~CMODE_H;
	    } else {
		newlock_off |= CMODE_H;
		newlock_on &= ~CMODE_H;
	    }
	    break;
#endif
	  case 'i':
	    if (add) {
		newlock_on |= CMODE_I;
		newlock_off &= ~CMODE_I;
	    } else {
		newlock_off |= CMODE_I;
		newlock_on &= ~CMODE_I;
	    }
	    break;
	  case 'k':
	    if (add) {
		if (!(s = strtok(NULL, " "))) {
		    notice(s_ChanServ, u->nick, "Parameter required for mode lock +k.");
		    return;
		}
		if (newlock_key)
		    free(newlock_key);
		newlock_key = sstrdup(s);
		newlock_off &= ~CMODE_K;
	    } else {
		if (newlock_key) {
		    free(newlock_key);
		    newlock_key = NULL;
		}
		newlock_off |= CMODE_K;
	    }
	    break;
	  case 'l':
	    if (add) {
		if (!(s = strtok(NULL, " "))) {
		    notice(s_ChanServ, u->nick, "Parameter required for mode lock +l.");
		    return;
		}
		if (atol(s) <= 0) {
		    notice(s_ChanServ, u->nick,
			"Parameter for mode lock +l must be a positive number.");
		    return;
		}
		newlock_limit = atol(s);
		newlock_off &= ~CMODE_L;
	    } else {
		newlock_limit = 0;
		newlock_off |= CMODE_L;
	    }
	    break;
	  case 'm':
	    if (add) {
		newlock_on |= CMODE_M;
		newlock_off &= ~CMODE_M;
	    } else {
		newlock_off |= CMODE_M;
		newlock_on &= ~CMODE_M;
	    }
	    break;
	  case 'n':
	    if (add) {
		newlock_on |= CMODE_N;
		newlock_off &= ~CMODE_N;
	    } else {
		newlock_off |= CMODE_N;
		newlock_on &= ~CMODE_N;
	    }
	    break;
	  case 'p':
	    if (add) {
		newlock_on |= CMODE_P;
		newlock_off &= ~CMODE_P;
	    } else {
		newlock_off |= CMODE_P;
		newlock_on &= ~CMODE_P;
	    }
	    break;
	  case 's':
	    if (add) {
		newlock_on |= CMODE_S;
		newlock_off &= ~CMODE_S;
	    } else {
		newlock_off |= CMODE_S;
		newlock_on &= ~CMODE_S;
	    }
	    break;
	  case 't':
	    if (add) {
		newlock_on |= CMODE_T;
		newlock_off &= ~CMODE_T;
	    } else {
		newlock_off |= CMODE_T;
		newlock_on &= ~CMODE_T;
	    }
	    break;
	  default:
	    notice(s_ChanServ, u->nick,
			"Channel mode character \2%c\2 is unknown and has been ignored.", c);
	    break;
	} /* switch */
    } /* while (*param) */

    /* Now that everything's okay, actually set the new mode lock. */
    newlock_on |= CMODE_R; // always means we are registered.
    newlock_off &= ~CMODE_R;
    ci->mlock_on = newlock_on;
    ci->mlock_off = newlock_off;
    ci->mlock_limit = newlock_limit;
    if (ci->mlock_key)
	free(ci->mlock_key);
    ci->mlock_key = newlock_key;

    /* Tell the user about it. */
    end = modebuf;
    *end = 0;
    if (ci->mlock_on)
      /* Just put all the modes into one var, non used have NULL or "" value, so it does not affect anything. */
	end += sprintf(end, "+%s%s%s%s%s%s%s%s%s%s%s",
#ifdef V_BAHAMUT
				(ci->mlock_on & CMODE_C) ? "c" : "",
#else
				"",
#endif
#ifdef V_UNREAL
				(ci->mlock_on & CMODE_H) ? "h" : "",
#else
				"",
#endif
				(ci->mlock_on & CMODE_I) ? "i" : "",
				(ci->mlock_key         ) ? "k" : "",
				(ci->mlock_limit       ) ? "l" : "",
				(ci->mlock_on & CMODE_M) ? "m" : "",
				(ci->mlock_on & CMODE_N) ? "n" : "",
				(ci->mlock_on & CMODE_P) ? "p" : "",
				(ci->mlock_on & CMODE_S) ? "s" : "",
				(ci->mlock_on & CMODE_R) ? "r" : "",
				(ci->mlock_on & CMODE_T) ? "t" : "");

    if (ci->mlock_off)
	end += sprintf(end, "-%s%s%s%s%s%s%s%s%s%s%s",
#ifdef V_BAHAMUT
				(ci->mlock_off & CMODE_C) ? "c" : "",
#else
				"",
#endif
#ifdef V_UNREAL
				(ci->mlock_off & CMODE_H) ? "h" : "",
#else
				"",
#endif
				(ci->mlock_off & CMODE_I) ? "i" : "",
				(ci->mlock_off & CMODE_K) ? "k" : "",
				(ci->mlock_off & CMODE_L) ? "l" : "",
				(ci->mlock_off & CMODE_M) ? "m" : "",
				(ci->mlock_off & CMODE_N) ? "n" : "",
				(ci->mlock_off & CMODE_P) ? "p" : "",
				(ci->mlock_off & CMODE_S) ? "s" : "",
				(ci->mlock_off & CMODE_R) ? "r" : "",
				(ci->mlock_off & CMODE_T) ? "t" : "");
    if (*modebuf) {
	notice(s_ChanServ, u->nick,
		"Mode lock for the channel \2%s\2 changed to \2%s\2.",
		ci->name, modebuf);
    } else {
	notice(s_ChanServ, u->nick,
		"Mode lock for the channel \2%s\2 removed.", ci->name);
    }

    /* Implement the new lock. */
    check_modes(ci->name);
}

/*************************************************************************/

static void cs_set_keeptopic(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_KEEPTOPIC;
	notice(s_ChanServ, u->nick, "Topic retention option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_KEEPTOPIC;
	notice(s_ChanServ, u->nick, "Topic retention option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s KEEPTOPIC [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET KEEPTOPIC\2.", s_ChanServ);
    }
}

/*************************************************************************/

static void cs_set_opguard(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_OPGUARD;
	notice(s_ChanServ, u->nick, "OpGuard option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_OPGUARD;
	notice(s_ChanServ, u->nick, "OpGuard option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s OPGUARD [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET OPGUARD\2.", s_ChanServ);
    }
}
/*************************************************************************/

static void cs_set_verbose(User *u, ChannelInfo *ci, char *param)
{
   if (!param) {
    if (ci->flags & CI_VERBOSE) {
     notice(s_ChanServ, u->nick, "VERBOSE option for \2%s\2 is \2ON\2", ci->name);
    } else {
     notice(s_ChanServ, u->nick, "VERBOSE option for \2%s\2 is \2OFF\2", ci->name);
    }
   } else {
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_VERBOSE;
	notice(s_ChanServ, u->nick, "VERBOSE option for the \2%s\2 is now set to \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_VERBOSE;
	notice(s_ChanServ, u->nick, "VERBOSE option for the \2%s\2 is now set to \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Usage: SET <channel> VERBOSE [ON|OFF]");
	notice(s_ChanServ, u->nick, "For help: /msg %s@%s HELP SET VERBOSE", s_ChanServ, server_name);
    }
  }
}

/*************************************************************************/

static void cs_set_topiclock(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_TOPICLOCK;
	notice(s_ChanServ, u->nick, "Topic lock option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_TOPICLOCK;
	notice(s_ChanServ, u->nick, "Topic lock option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s TOPICLOCK [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET TOPICLOCK\2.", s_ChanServ);
    }
}

/*************************************************************************/

static void cs_set_ircop(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_IRCOP;
	notice(s_ChanServ, u->nick, "IRC Operators option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_IRCOP;
	notice(s_ChanServ, u->nick, "IRC Operators option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s IRCOP [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET IRCOP\2.", s_ChanServ);
    }
}

/*************************************************************************/

static void cs_set_admin(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_ADMIN;
	notice(s_ChanServ, u->nick, "Services Adminstrators option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_ADMIN;
	notice(s_ChanServ, u->nick, "Services Adminstrators option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s ADMIN [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET ADMIN\2.", s_ChanServ);
    }
}

/*************************************************************************/

static void cs_set_private(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_PRIVATE;
	notice(s_ChanServ, u->nick, "Private option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_PRIVATE;
	notice(s_ChanServ, u->nick, "Private option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s PRIVATE [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET PRIVATE\2.", s_ChanServ);
    }
}

/*************************************************************************/

static void cs_set_secureops(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_SECUREOPS;
	notice(s_ChanServ, u->nick, "Secure ops option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_SECUREOPS;
	notice(s_ChanServ, u->nick, "Secure ops option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s SECUREOPS [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET SECUREOPS\2.", s_ChanServ);
    }
}

/*************************************************************************/

static void cs_set_restricted(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_RESTRICTED;
	notice(s_ChanServ, u->nick, "Restricted access option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_RESTRICTED;
	notice(s_ChanServ, u->nick, "Restricted access option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s RESTRICTED [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET RESTRICTED\2.", s_ChanServ);
    }
}

/*************************************************************************/

static void cs_set_secure(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_SECURE;
	notice(s_ChanServ, u->nick, "Secure option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_SECURE;
	notice(s_ChanServ, u->nick, "Secure option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s SECURE [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET SECURE\2.", s_ChanServ);
    }
}

/*************************************************************************/

static void cs_set_protected(User *u, ChannelInfo *ci, char *param)
{
    if (stricmp(param, "ON") == 0) {
	ci->flags |= CI_PROTECTED;
	notice(s_ChanServ, u->nick, "Protected option for the channel \2%s\2 is now \2ON\2.", ci->name);
    } else if (stricmp(param, "OFF") == 0) {
	ci->flags &= ~CI_PROTECTED;
	notice(s_ChanServ, u->nick, "Protected option for the channel \2%s\2 is now \2OFF\2.", ci->name);
    } else {
	notice(s_ChanServ, u->nick, "Syntax: SET %s PROTECTED [ON|OFF]", ci->name);
	notice(s_ChanServ, u->nick, "For more information please type \2/msg %s HELP SET PROTECTED\2.", s_ChanServ);
    }
}

/*************************************************************************/

/* cs_aop()
 * Add a nick to the channel AutoOperator access list. */

static void cs_aop(const char *source)
{
   char *chan = strtok(NULL, " ");
   char *cmd = strtok(NULL, " ");
   char *nick;
   ChannelInfo *ci;
   NickInfo *ni;
   User *u;
   ChanAccess *access;
   short aop_level = ret_level(CA_AUTOOP);
   short level = 0, ulev;
   int i, ii = 0;

   if (!chan || !cmd) {
	notice(s_ChanServ, source, "Syntax: AOP %s [ADD|DEL|LIST] <nick>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP AOP\2.",
			s_ChanServ);
   } else if (!(ci = cs_findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
   } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
      log("%s: ACCESS command from nonexistent user %s", s_ChanServ, source);
   } else if (stricmp(cmd, "ADD") == 0) {
	   nick = strtok(NULL, " ");
	if (!nick) {
		notice(s_ChanServ, source, "Syntax: AOP %s ADD <nick>",
			(chan && *chan == '#') ? chan : "<channel>");
		notice(s_ChanServ, source, "For more information please type \2/msg %s HELP AOP\2.",
			s_ChanServ);
		return;
 	}
	if (!check_access(u, ci, CA_ACCESS_CHANGE)) {
         notice(s_ChanServ, source, "Access denied.");
         return;
      }
      level = aop_level;
      ulev = get_access(u, ci);
      /* Sop add Aop, nothing higher. */
      if (level >= ulev) {
         notice(s_ChanServ, source, "Access denied.");
         return;
      }
      if (!(ni = findnick(nick))) {
         notice(s_ChanServ, source, "Channel AOp lists may only contain registered nicks, "
            "not nick!user@host masks.");
         return;
      }
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (access->in_use && stricmp(access->name, nick) == 0) {
		/* Don't allow lowering from a level >= ulev */
		if (access->level >= ulev) {
		    notice(s_ChanServ, source, "Access denied.");
		    return;
		}
            if ((ni->flags & NI_NOOPS)) {
		    notice(s_ChanServ, source, "\2%s\2 does not want to be added to channel operator lists at this time.", ni->nick);
		    return;
		}
		if (access->level == level) {
		    notice(s_ChanServ, source, "\2%s\2 is already on the channel \2%s\2 AOp list.", access->name, chan);
		    return;
		}
		if (access->level == ret_level(CA_AKICK)) {
                    if (ci->flags & CI_VERBOSE) {
                    send_cmd(NULL, ":%s NOTICE @%s :[VERBOSE] \2%s\2!%s@%s => aop %s %s %s", s_ChanServ, chan, source, u->username, u->host, chan, cmd, nick);
                    }
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from SOp to AOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
        	if (access->level == ret_level(CA_AUTOVOICE)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from VOp to AOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
#ifdef V_UNREAL
		if (access->level == ret_level(CA_AUTOHALFOP)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from HOp to AOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
#endif
	    }
	}
	for (i = 0; i < ci->accesscount; i++) {
	    if (!ci->access[i].in_use)
		break;
	}
	if (i == ci->accesscount) {
	    ci->accesscount++;
	    ci->access =
		srealloc(ci->access, sizeof(ChanAccess) * ci->accesscount);
	}
	if (ci->flags & CI_VERBOSE) {
            send_cmd(NULL, ":%s NOTICE @%s :[VERBOSE] \2%s\2!%s@%s => aop %s %s %s", s_ChanServ, chan, source, u->username, u->host, chan, cmd, nick);
        }
	access = &ci->access[i];
	access->name = sstrdup(ni->nick);
	access->in_use = 1;
	access->level = level;
	notice(s_ChanServ, source, "\2%s\2 added to the channel \2%s\2 AOp list.", access->name, chan, level);
    } else if (stricmp(cmd, "DEL") == 0) {
	   nick = strtok(NULL, " ");
	if (!nick) {
		notice(s_ChanServ, source, "Syntax: AOP %s DEL <nick>",
			(chan && *chan == '#') ? chan : "<channel>");
		notice(s_ChanServ, source, "For more information please type \2/msg %s HELP AOP\2.",
			s_ChanServ);
		return;
	}
	if (!check_access(u, ci, CA_ACCESS_CHANGE)) {
         notice(s_ChanServ, source, "Acces denied.");
		return;
      }
       /* Special case: is it a number?  Only do search if it isn't. */
       if (strspn(nick, "1234567890") == strlen(nick) && (i = atoi(nick)) > 0 && i <= ci->accesscount) {
          i--;
          if (!ci->access[i].in_use) {
             notice(s_ChanServ, source, "No such entry (#%d) on the channel \2%s\2 AOp list.", i + 1, chan);
             return;
	    }
       } else {
	    if (!(ni = findnick(nick))) {
		notice(s_ChanServ, source, "Nick \2%s\2 is not registered.", nick);
		return;
	    }
	    for (i = 0; i < ci->accesscount; i++) {
		if (ci->access[i].in_use && stricmp(ci->access[i].name, nick) == 0)
		    break;
	    }
	    if (i == ci->accesscount) {
		notice(s_ChanServ, source, "\2%s\2 not found on the channel \2%s\2 AOp list.", nick, chan);
		return;
	    }
	}
	access = &ci->access[i];
	if (get_access(u, ci) <= access->level) {
	    notice(s_ChanServ, source, "Access denied.");
		return;
	} else {
          if (ci->flags & CI_VERBOSE) {
           send_cmd(NULL, ":%s NOTICE @%s :[VERBOSE] \2%s\2!%s@%s => aop %s %s %s", s_ChanServ, chan, source, u->username, u->host, chan, cmd, nick);
          }
	    notice(s_ChanServ, source,
		"\2%s\2 deleted from the channel \2%s\2 AOp list.", access->name, chan);
	    free(access->name);
	    access->name = NULL;
	    access->in_use = 0;
	}
    } else if (stricmp(cmd, "LIST") == 0) {
	if (!check_access(u, ci, CA_ACCESS_LIST) && !is_oper(source)) {
         notice(s_ChanServ, source, "Access denied.");
		return;
      }
	notice(s_ChanServ, source, "Current AOp list for %s:", chan);
	notice(s_ChanServ, source, "Num Nick");
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (!access->in_use)
		continue;
	    if ((ni = findnick(access->name)) && (access->level == aop_level)) {
		ii++;
             notice(s_ChanServ, source, "%3d %s (%s)", 
                ii, access->name, ni->last_usermask);
          }
	}
   } else {
	notice(s_ChanServ, source, "Syntax: AOP %s [ADD|DEL|LIST] <nick>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP AOP\2.",
			s_ChanServ);
   }

}

static void cs_sop(const char *source)
{
   char *chan = strtok(NULL, " ");
   char *cmd = strtok(NULL, " ");
   char *nick;
   ChannelInfo *ci;
   NickInfo *ni;
   User *u;
   ChanAccess *access;
   short sop_level = ret_level(CA_AKICK);
   short level = 0, ulev;
   int i, ii = 0;

   if (!chan || !cmd) {
	notice(s_ChanServ, source, "Syntax: SOP %s [ADD|DEL|LIST] <nick>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP SOP\2.",
			s_ChanServ);
   } else if (!(ci = cs_findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
   } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
      log("%s: ACCESS command from nonexistent user %s", s_ChanServ, source);
   } else if (stricmp(cmd, "ADD") == 0) {
	   nick = strtok(NULL, " ");
	if (!nick) {
		notice(s_ChanServ, source, "Syntax: SOP %s ADD <nick>",
			(chan && *chan == '#') ? chan : "<channel>");
		notice(s_ChanServ, source, "For more information please type \2/msg %s HELP SOP\2.",
			s_ChanServ);
		return;
 	}
	if (!check_access(u, ci, CA_ACCESS_CHANGE)) {
         notice(s_ChanServ, source, "Access denied.");
		return;
      }
      level = sop_level;
      ulev = get_access(u, ci);
      /* Sop can add Aop, nothing higher. */
      if (level >= ulev) {
         notice(s_ChanServ, source, "Access denied.");
         return;
      }
      if (!(ni = findnick(nick))) {
         notice(s_ChanServ, source, "Channel SOp lists may only contain registered nicks, "
            "not nick!user@host masks.");
         return;
      }
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (access->in_use && stricmp(access->name, nick) == 0) {
		/* Don't allow lowering from a level >= ulev */
		if (access->level >= ulev) {
		    notice(s_ChanServ, source, "Access denied.");
		    return;
		}
            if ((ni->flags & NI_NOOPS)) {
		    notice(s_ChanServ, source, "\2%s\2 does not want to be added to channel operator lists at this time.", ni->nick);
		    return;
		}
		if (access->level == level) {
		    notice(s_ChanServ, source, "\2%s\2 is already on the channel \2%s\2 SOp list.", access->name, chan);
		    return;
		}
		if (access->level == ret_level(CA_AUTOOP)) {
                if (ci->flags & CI_VERBOSE) {
                    send_cmd(NULL, ":%s NOTICE @%s :[VERBOSE] \2%s\2!%s@%s => sop %s %s %s", s_ChanServ, chan, source, u->username, u->host, chan, cmd, nick);
                    }
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from AOp to SOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
		if (access->level == ret_level(CA_AUTOVOICE)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from VOp to SOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
#ifdef V_UNREAL
		if (access->level == ret_level(CA_AUTOHALFOP)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from HOp to SOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
#endif
	    }
	}
	for (i = 0; i < ci->accesscount; i++) {
	    if (!ci->access[i].in_use)
		break;
	}
	if (i == ci->accesscount) {
	    ci->accesscount++;
	    ci->access =
		srealloc(ci->access, sizeof(ChanAccess) * ci->accesscount);
	}
        if (ci->flags & CI_VERBOSE) {
           send_cmd(NULL, ":%s NOTICE @%s :[VERBOSE] \2%s\2!%s@%s => sop %s %s %s", s_ChanServ, chan, source, u->username, u->host, chan, cmd, nick);
        }
	access = &ci->access[i];
	access->name = sstrdup(ni->nick);
	access->in_use = 1;
	access->level = level;
	notice(s_ChanServ, source, "\2%s\2 added to the channel \2%s\2 SOp list.", access->name, chan, level);
    } else if (stricmp(cmd, "DEL") == 0) {
	   nick = strtok(NULL, " ");
	if (!nick) {
		notice(s_ChanServ, source, "Syntax: SOP %s DEL <nick>",
			(chan && *chan == '#') ? chan : "<channel>");
		notice(s_ChanServ, source, "For more information please type \2/msg %s HELP SOP\2.",
			s_ChanServ);
		return;
 	}
	if (!check_access(u, ci, CA_ACCESS_CHANGE)) {
         notice(s_ChanServ, source, "Access denied.");
		return;
      }
       /* Special case: is it a number?  Only do search if it isn't. */
       if (strspn(nick, "1234567890") == strlen(nick) && (i = atoi(nick)) > 0 && i <= ci->accesscount) {
          i--;
          if (!ci->access[i].in_use) {
             notice(s_ChanServ, source, "No such entry (#%d) on the channel \2%s\2 SOp list.", i+1, chan);
             return;
	    }
       } else {
	    if (!(ni = findnick(nick))) {
		notice(s_ChanServ, source, "Nick \2%s\2 is not registered.", nick);
		return;
	    }
	    for (i = 0; i < ci->accesscount; i++) {
		if (ci->access[i].in_use && stricmp(ci->access[i].name, nick) == 0)
		    break;
	    }
	    if (i == ci->accesscount) {
		notice(s_ChanServ, source, "\2%s\2 not found on the channel \2%s\2 SOp list.", nick, chan);
		return;
	    }
	}
	access = &ci->access[i];
	if (get_access(u, ci) <= access->level) {
	    notice(s_ChanServ, source, "Access denied.");
		return;
	} else {
        if (ci->flags & CI_VERBOSE) {
            send_cmd(NULL, ":%s NOTICE @%s :[VERBOSE] \2%s\2!%s@%s => sop %s %s %s", s_ChanServ, chan, source, u->username, u->host, chan, cmd, nick);
            }
	    notice(s_ChanServ, source,
		"\2%s\2 deleted from the channel \2%s\2 SOp list.", access->name, chan);
	    free(access->name);
	    access->name = NULL;
	    access->in_use = 0;
	}
    } else if (stricmp(cmd, "LIST") == 0) {
	if (!check_access(u, ci, CA_ACCESS_LIST) && !is_oper(source)) {
         notice(s_ChanServ, source, "Access denied.");
		return;
      }
	notice(s_ChanServ, source, "Current SOp list for %s:", chan);
	notice(s_ChanServ, source, "Num Nick");
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (!access->in_use)
		continue;
	    if ((ni = findnick(access->name)) && (access->level == sop_level)) {
		ii++;
             notice(s_ChanServ, source, "%3d %s (%s)", 
                ii, access->name, ni->last_usermask);
          }
	}
   } else {
	notice(s_ChanServ, source, "Syntax: SOP %s [ADD|DEL|LIST] <nick>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP SOP\2.",
			s_ChanServ);
   }
}

static void cs_vop(const char *source)
{
   char *chan = strtok(NULL, " ");
   char *cmd = strtok(NULL, " ");
   char *nick;
   ChannelInfo *ci;
   NickInfo *ni;
   User *u;
   ChanAccess *access;
   short vop_level = ret_level(CA_AUTOVOICE);
   short level = 0, ulev;
   int i, ii = 0;

   crash_id = 196;

   if (!chan || !cmd) {
	notice(s_ChanServ, source, "Syntax: VOP %s [ADD|DEL|LIST] <nick>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP VOP\2.",
			s_ChanServ);
   } else if (!(ci = cs_findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
   } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
      log("%s: ACCESS command from nonexistent user %s", s_ChanServ, source);
   } else if (stricmp(cmd, "ADD") == 0) {
	   nick = strtok(NULL, " ");
	if (!nick) {
		notice(s_ChanServ, source, "Syntax: VOP %s ADD <nick>",
			(chan && *chan == '#') ? chan : "<channel>");
		notice(s_ChanServ, source, "For more information please type \2/msg %s HELP VOP\2.",
			s_ChanServ);
		return;
 	}
	if (!check_access(u, ci, CA_ACCESS_CHANGE)) {
         notice(s_ChanServ, source, "Access denied.");
		return;
      }
      level = vop_level;
      ulev = get_access(u, ci);
      /* Sop can add Aop, nothing higher. */
      if (level >= ulev) {
         notice(s_ChanServ, source, "Access denied.");
         return;
      }
      if (!(ni = findnick(nick))) {
         notice(s_ChanServ, source, "Channel VOp lists may only contain registered nicks, "
            "not nick!user@host masks.");
         return;
      }
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (access->in_use && stricmp(access->name, nick) == 0) {
		/* Don't allow lowering from a level >= ulev */
		if (access->level >= ulev) {
		    notice(s_ChanServ, source, "Access denied.");
		    return;
		}
            if ((ni->flags & NI_NOOPS)) {
		    notice(s_ChanServ, source, "\2%s\2 does not want to be added to channel voice lists at this time.", ni->nick);
		    return;
		}
		if (access->level == level) {
		    notice(s_ChanServ, source, "\2%s\2 is already on the channel \2%s\2 VOp list.", access->name, chan);
		    return;
		}
		if (access->level == ret_level(CA_AUTOOP)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from AOp to VOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
		if (access->level == ret_level(CA_AKICK)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from SOp to VOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
#ifdef V_UNREAL
		if (access->level == ret_level(CA_AUTOHALFOP)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from HOp to VOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
#endif
	    }
	}
	for (i = 0; i < ci->accesscount; i++) {
	    if (!ci->access[i].in_use)
		break;
	}
	if (i == ci->accesscount) {
	    ci->accesscount++;
	    ci->access =
		srealloc(ci->access, sizeof(ChanAccess) * ci->accesscount);
	}
	access = &ci->access[i];
	access->name = sstrdup(ni->nick);
	access->in_use = 1;
	access->level = level;
	notice(s_ChanServ, source, "\2%s\2 added to the channel \2%s\2 VOp list.", access->name, chan, level);
    } else if (stricmp(cmd, "DEL") == 0) {
	   nick = strtok(NULL, " ");
	if (!nick) {
		notice(s_ChanServ, source, "Syntax: VOP %s DEL <nick>",
			(chan && *chan == '#') ? chan : "<channel>");
		notice(s_ChanServ, source, "For more information please type \2/msg %s HELP VOP\2.",
			s_ChanServ);
		return;
 	}
	if (!check_access(u, ci, CA_ACCESS_CHANGE)) {
         notice(s_ChanServ, source, "Access denied.");
		return;
      }
       /* Special case: is it a number?  Only do search if it isn't. */
       if (strspn(nick, "1234567890") == strlen(nick) && (i = atoi(nick)) > 0 && i <= ci->accesscount) {
          i--;
          if (!ci->access[i].in_use) {
             notice(s_ChanServ, source, "No such entry (#%d) on the channel \2%s\2 VOp list.", i+1, chan);
             return;
	    }
       } else {
	    if (!(ni = findnick(nick))) {
		notice(s_ChanServ, source, "Nick \2%s\2 is not registered.", nick);
		return;
	    }
	    for (i = 0; i < ci->accesscount; i++) {
		if (ci->access[i].in_use && stricmp(ci->access[i].name, nick) == 0)
		    break;
	    }
	    if (i == ci->accesscount) {
		notice(s_ChanServ, source, "\2%s\2 not found on the channel \2%s\2 VOp list.", nick, chan);
		return;
	    }
	}
	access = &ci->access[i];
	if (get_access(u, ci) <= access->level) {
	    notice(s_ChanServ, source, "Access denied.");
		return;
	} else {
	    notice(s_ChanServ, source,
		"\2%s\2 deleted from the channel \2%s\2 VOp list.", access->name, chan);
	    free(access->name);
	    access->name = NULL;
	    access->in_use = 0;
	}
    } else if (stricmp(cmd, "LIST") == 0) {
	if (!check_access(u, ci, CA_ACCESS_LIST) && !is_oper(source)) {
         notice(s_ChanServ, source, "Access denied.");
		return;
      }
	notice(s_ChanServ, source, "Current VOp list for %s:", chan);
	notice(s_ChanServ, source, "Num Nick");
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (!access->in_use)
		continue;
	    if ((ni = findnick(access->name)) && (access->level == vop_level)) {
		ii++;
             notice(s_ChanServ, source, "%3d %s (%s)", 
                ii, access->name, ni->last_usermask);
          }
	}
   } else {
	notice(s_ChanServ, source, "Syntax: VOP %s [ADD|DEL|LIST] <nick>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP VOP\2.",
			s_ChanServ);
   }
}

/*************************************************************************/

#ifdef V_UNREAL
/* cs_hop()
 * Add a nick to the channel AutoHalfOperator access list. */

static void cs_hop(const char *source)
{
   char *chan = strtok(NULL, " ");
   char *cmd = strtok(NULL, " ");
   char *nick;
   ChannelInfo *ci;
   NickInfo *ni;
   User *u;
   ChanAccess *access;
   short hop_level = ret_level(CA_AUTOHALFOP);
   short level = 0, ulev;
   int i, ii = 0;

   if (!chan || !cmd) {
	notice(s_ChanServ, source, "Syntax: HOP %s [ADD|DEL|LIST] <nick>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP HOP\2.",
			s_ChanServ);
   } else if (!(ci = cs_findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
   } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
      log("%s: ACCESS command from nonexistent user %s", s_ChanServ, source);
   } else if (stricmp(cmd, "ADD") == 0) {
	   nick = strtok(NULL, " ");
	if (!nick) {
		notice(s_ChanServ, source, "Syntax: HOP %s ADD <nick>",
			(chan && *chan == '#') ? chan : "<channel>");
		notice(s_ChanServ, source, "For more information please type \2/msg %s HELP HOP\2.",
			s_ChanServ);
		return;
 	}
	if (!check_access(u, ci, CA_ACCESS_CHANGE)) {
         notice(s_ChanServ, source, "Access denied.");
         return;
      }
      level = hop_level;
      ulev = get_access(u, ci);
      /* Sop add Aop, nothing higher. */
      if (level >= ulev) {
         notice(s_ChanServ, source, "Access denied.");
         return;
      }
      if (!(ni = findnick(nick))) {
         notice(s_ChanServ, source, "Channel HOp lists may only contain registered nicks, "
            "not nick!user@host masks.");
         return;
      }
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (access->in_use && stricmp(access->name, nick) == 0) {
		/* Don't allow lowering from a level >= ulev */
		if (access->level >= ulev) {
		    notice(s_ChanServ, source, "Access denied.");
		    return;
		}
            if ((ni->flags & NI_NOOPS)) {
		    notice(s_ChanServ, source, "\2%s\2 does not want to be added to channel operator lists at this time.", ni->nick);
		    return;
		}
		if (access->level == level) {
		    notice(s_ChanServ, source, "\2%s\2 is already on the channel \2%s\2 HOp list.", access->name, chan);
		    return;
		}
		if (access->level == ret_level(CA_AKICK)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from SOp to HOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
		if (access->level == ret_level(CA_AUTOVOICE)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from VOp to HOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
		if (access->level == ret_level(CA_AUTOOP)) {
		    access->level = level;
		    notice(s_ChanServ, source, "\2%s\2 has been changed from AOp to HOp on the channel \2%s\2.", access->name, chan);
		    return;
		}
	    }
	}
	for (i = 0; i < ci->accesscount; i++) {
	    if (!ci->access[i].in_use)
		break;
	}
	if (i == ci->accesscount) {
	    ci->accesscount++;
	    ci->access =
		srealloc(ci->access, sizeof(ChanAccess) * ci->accesscount);
	}
	access = &ci->access[i];
	access->name = sstrdup(ni->nick);
	access->in_use = 1;
	access->level = level;
	notice(s_ChanServ, source, "\2%s\2 added to the channel \2%s\2 HOp list.", access->name, chan, level);
    } else if (stricmp(cmd, "DEL") == 0) {
	   nick = strtok(NULL, " ");
	if (!nick) {
		notice(s_ChanServ, source, "Syntax: HOP %s DEL <nick>",
			(chan && *chan == '#') ? chan : "<channel>");
		notice(s_ChanServ, source, "For more information please type \2/msg %s HELP HOP\2.",
			s_ChanServ);
		return;
	}
	if (!check_access(u, ci, CA_ACCESS_CHANGE)) {
         notice(s_ChanServ, source, "Acces denied.");
		return;
      }
       /* Special case: is it a number?  Only do search if it isn't. */
       if (strspn(nick, "1234567890") == strlen(nick) && (i = atoi(nick)) > 0 && i <= ci->accesscount) {
          i--;
          if (!ci->access[i].in_use) {
             notice(s_ChanServ, source, "No such entry (#%d) on the channel \2%s\2 HOp list.", i + 1, chan);
             return;
	    }
       } else {
	    if (!(ni = findnick(nick))) {
		notice(s_ChanServ, source, "Nick \2%s\2 is not registered.", nick);
		return;
	    }
	    for (i = 0; i < ci->accesscount; i++) {
		if (ci->access[i].in_use && stricmp(ci->access[i].name, nick) == 0)
		    break;
	    }
	    if (i == ci->accesscount) {
		notice(s_ChanServ, source, "\2%s\2 not found on the channel \2%s\2 HOp list.", nick, chan);
		return;
	    }
	}
	access = &ci->access[i];
	if (get_access(u, ci) <= access->level) {
	    notice(s_ChanServ, source, "Access denied.");
		return;
	} else {
	    notice(s_ChanServ, source,
		"\2%s\2 deleted from the channel \2%s\2 HOp list.", access->name, chan);
	    free(access->name);
	    access->name = NULL;
	    access->in_use = 0;
	}
    } else if (stricmp(cmd, "LIST") == 0) {
	if (!check_access(u, ci, CA_ACCESS_LIST) && !is_oper(source)) {
         notice(s_ChanServ, source, "Access denied.");
		return;
      }
	notice(s_ChanServ, source, "Current HOp list for %s:", chan);
	notice(s_ChanServ, source, "Num Nick");
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (!access->in_use)
		continue;
	    if ((ni = findnick(access->name)) && (hop_level == access->level)) {
		ii++;
             notice(s_ChanServ, source, "%3d %s (%s)", 
                ii, access->name, ni->last_usermask);
          }
	}
   } else {
	notice(s_ChanServ, source, "Syntax: HOP %s [ADD|DEL|LIST] <nick>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP HOP\2.",
			s_ChanServ);
   }

}
#endif

/*************************************************************************/

static void cs_why(const char *source)
{
   char *chan = strtok(NULL, " ");
   char *nick = strtok(NULL, " ");
   ChannelInfo *ci;
   User *u, *uu;
   NickInfo *ni;
   int ulev = 0, uulev = 0;
   char *anick;

   if (!chan || !nick) {
      notice(s_ChanServ, source, "Syntax: WHY <channel> <nick>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP WHY\2.", s_ChanServ);
   } else if (!(ci = cs_findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 is not registered", chan);
   } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
   } else if ((uu = finduser(nick))) {
      ulev = get_access(u, ci);
      anick = get_why_access(uu, ci);
      uulev = get_access(uu, ci);
      
      if (ulev < ret_level(CA_AUTOOP) && !is_oper(source)) {
         notice(s_ChanServ, source, "Access denied.");
      } else if (is_services_root(uu->nick) && uulev <= 0) {
            notice(s_ChanServ, source, "\2%s\2 has special access to the channel \2%s\2 because they "
               "have identified as a Services Root.", 
                  uu->nick, chan);
      } else {
         if (uulev <= 0) {
	      notice(s_ChanServ, source, "\2%s\2 does not have access to the channel \2%s\2.",
	         nick, chan);
         } else if (uulev == ACCESS_FOUNDER) {
            notice(s_ChanServ, source, "\2%s\2 has access to the channel \2%s\2 because they "
               "have identified themself as founder.", 
                  uu->nick, chan);
         } else if (uulev == ret_level(CA_AUTOVOICE)) {
            notice(s_ChanServ, source, "\2%s\2 has access to the channel \2%s\2 because they "
               "have identified for the nick \2%s\2 which is on the channel VOp list.", 
                  uu->nick, chan, anick);
         } else if (uulev == ret_level(CA_AUTOOP)) {
            notice(s_ChanServ, source, "\2%s\2 has access to the channel \2%s\2 because they "
               "have identified for the nick \2%s\2 which is on the channel AOp list.", 
                  uu->nick, chan, anick);
         } else if (uulev == ret_level(CA_AKICK)) {
            notice(s_ChanServ, source, "\2%s\2 has access to the channel \2%s\2 because they "
               "have identified for the nick \2%s\2 which is on the channel SOp list.", 
                  uu->nick, chan, anick);
#ifdef V_UNREAL
         } else if (uulev == ret_level(CA_AUTOHALFOP)) {
            notice(s_ChanServ, source, "\2%s\2 has access to the channel \2%s\2 because they "
               "have identified for the nick \2%s\2 which is on the channel HOp list.", 
                  uu->nick, chan, anick);
#endif
         } else {
            notice(s_ChanServ, source, "\2%s\2 has access to the channel \2%s\2 because they "
               "have identified for the nick \2%s\2 which is on the channel access list at level %d.",
                  uu->nick, chan, anick, uulev);
         }
      }

   } else {
      notice(s_ChanServ, source, "\2%s\2 does not have access to the channel \2%s\2.",
         nick, chan);
   }
}

/*************************************************************************/

static void cs_access(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *cmd  = strtok(NULL, " ");
    char *nick = strtok(NULL, " ");
    char *s    = strtok(NULL, " ");
    ChannelInfo *ci;
    NickInfo *ni;
    User *u;
    short level = 0, ulev;
    int i;
    int is_list = (cmd && stricmp(cmd, "LIST") == 0);
    ChanAccess *access;

    /* If LIST, we don't *require* any parameters, but we can take any.
     * If DEL, we require a nick and no level.
     * Else (ADD), we require a level (which implies a nick). */
    if (!cmd || (is_list ? 0 :
			(stricmp(cmd,"DEL")==0) ? (!nick || s) : !s)) {
	notice(s_ChanServ, source,
		"Syntax: ACCESS %s [ADD|DEL|LIST] <nick> <level>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP ACCESS\2.",
			s_ChanServ);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);

    } else if (!(u = finduser(source))) {

	notice(s_ChanServ, source, "Access denied.");
	log("%s: ACCESS command from nonexistent user %s", s_ChanServ, source);

    } else if (((is_list && !check_access(u, ci, CA_ACCESS_LIST))
             || (!is_list && !check_access(u, ci, CA_ACCESS_CHANGE)))
            && !(is_list && is_oper(source)))
    {
	notice(s_ChanServ, source, "Access denied.");

    } else if (stricmp(cmd, "ADD") == 0) {
	level = atoi(s);
	ulev = get_access(u, ci);
	if (level >= ulev) {
	    notice(s_ChanServ, source, "Access denied.");
	    return;
	}
	if (level == 0) {
	    notice(s_ChanServ, source, "Access level must be non-zero.");
	    return;
	} else if (level <= ACCESS_INVALID || level >= ACCESS_FOUNDER) {
	    notice(s_ChanServ, source,
			"Access level must be between %d and %d inclusive.",
			ACCESS_INVALID+1, ACCESS_FOUNDER-1);
	    return;
	}
	ni = findnick(nick);
	if (!ni) {
	    notice(s_ChanServ, source,
		"Channel access lists may only contain registered nicks, "
		"not nick!user@host masks.");
	    return;
	}
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (access->in_use && stricmp(access->name, nick) == 0) {
		/* Don't allow lowering from a level >= ulev */
		if (access->level >= ulev) {
		    notice(s_ChanServ, source, "Access denied.");
		    return;
		}
            if ((ni->flags & NI_NOOPS)) {
		    notice(s_ChanServ, source, "\2%s\2 does not want to be added to channel operator lists at this time.", ni->nick);
		    return;
		}
		if (access->level == level) {
		    notice(s_ChanServ, source,
			"Access level for \2%s\2 on the channel \2%s\2 unchanged from \2%d\2.",
			access->name, chan, level);
		    return;
		}
		access->level = level;
		notice(s_ChanServ, source,
			"Access level for \2%s\2 on the channel \2%s\2 changed to \2%d\2.",
			access->name, chan, level);
		return;
	    }
	}
	for (i = 0; i < ci->accesscount; i++) {
	    if (!ci->access[i].in_use)
		break;
	}
	if (i == ci->accesscount) {
	    ci->accesscount++;
	    ci->access =
		srealloc(ci->access, sizeof(ChanAccess) * ci->accesscount);
	}
	access = &ci->access[i];
	access->name = sstrdup(ni->nick);
	access->in_use = 1;
	access->level = level;
	notice(s_ChanServ, source,
		"\2%s\2 added to the channel \2%s\2 access list at level \2%d\2.",
		access->name, chan, level);

    } else if (stricmp(cmd, "DEL") == 0) {

	/* Special case: is it a number?  Only do search if it isn't. */
	if (strspn(nick, "1234567890") == strlen(nick) &&
				(i = atoi(nick)) > 0 && i <= ci->accesscount) {
	    i--;
	    if (!ci->access[i].in_use) {
		notice(s_ChanServ, source,
			"No such entry (#%d) on the channel \2%s\2 access list.", i+1, chan);
		return;
	    }
	} else {
	    ni = findnick(nick);
	    if (!ni) {
		notice(s_ChanServ, source,
			"Nick \2%s\2 is not registered.", nick);
		return;
	    }
	    for (i = 0; i < ci->accesscount; i++) {
		if (ci->access[i].in_use &&
		                  stricmp(ci->access[i].name, nick) == 0)
		    break;
	    }
	    if (i == ci->accesscount) {
		notice(s_ChanServ, source,
			"\2%s\2 not found on the channel \2%s\2 access list.", nick, chan);
		return;
	    }
	}
	access = &ci->access[i];
	if (get_access(u, ci) <= access->level) {
	    notice(s_ChanServ, source, "Access denied.");
	} else {
	    notice(s_ChanServ, source,
		"\2%s\2 deleted from the channel \2%s\2 access list.", access->name, chan);
	    free(access->name);
	    access->name = NULL;
	    access->in_use = 0;
	}

    } else if (stricmp(cmd, "LIST") == 0) {

	notice(s_ChanServ, source, "Current access list for %s:", chan);
	notice(s_ChanServ, source, "Num Level Nick");
	for (access = ci->access, i = 0; i < ci->accesscount; access++, i++) {
	    if (!access->in_use)
		continue;
	    if ((nick && !match_wild(nick, access->name)) || (level && level != access->level))
		continue;
	    if (ni = findnick(access->name))
		s = ni->last_usermask;
	    else
		s = NULL;
	    notice(s_ChanServ, source, "%3d %5d %s%s%s%s", i + 1, access->level, access->name,
			s ? " (" : "", s ? s : "", s ? ")" : "");
	}

    } else {

	notice(s_ChanServ, source,
		"Syntax: ACCESS %s [ADD|DEL|LIST] <nick> <level>",
		*chan == '#' ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP ACCESS\2.",
			s_ChanServ);

    }
}

/*************************************************************************/

static void cs_akick(const char *source)
{
    char *chan   = strtok(NULL, " ");
    char *cmd    = strtok(NULL, " ");
    char *mask   = strtok(NULL, " ");
    char *reason = strtok(NULL, "");
    char *t;
    ChannelInfo *ci;
    NickInfo *ni;
    User *u;
    int i;
    AutoKick *akick;

    if (!cmd || (stricmp(cmd, "LIST") != 0 && !mask)) {

	notice(s_ChanServ, source,
		"Syntax: AKICK %s [ADD|DEL|LIST] <user@host> <\37nick\37>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP AKICK\2.",
			s_ChanServ);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);

    } else if ((!(u = finduser(source)) || !check_access(u, ci, CA_AKICK))
	       && !(stricmp(cmd, "LIST") == 0 && is_oper(source))) {

	notice(s_ChanServ, source, "Access denied.");

    } else if (stricmp(cmd, "ADD") == 0) {

	NickInfo *ni = findnick(mask);
	char *nick, *user, *host;

	if (!ni) {
	    if (debug >= 3) {
		log("debug: %s: cs_akick(add): splitting mask %s",
			s_ChanServ, mask);
	    }
	    split_usermask(mask, &nick, &user, &host);
	    mask = smalloc(strlen(nick)+strlen(user)+strlen(host)+3);
	    sprintf(mask, "%s!%s@%s", nick, user, host);
	    free(nick);
	    free(user);
	    free(host);
	}

	if (debug >= 3) {
	    log("debug: %s: cs_akick(add): searching for akick matching %s [%p]",
			s_ChanServ, mask, mask);
	}
	for (akick = ci->akick, i = 0; i < ci->akickcount; akick++, i++) {
	    if (akick->is_nick < 0)
		continue;
	    if ((akick->is_nick ? stricmp(akick->name,mask) :
	                          strcmp(akick->name,mask)) == 0) {
		notice(s_ChanServ, source,
			"\2%s\2 already exists on %s autokick list.",
			akick->name, chan);
		return;
	    }
	}

	if (debug >= 3) {
	    log("debug: %s: cs_akick(add): searching for free akick",
			s_ChanServ);
	}
	for (i = 0; i < ci->akickcount; i++) {
	    if (ci->akick[i].is_nick < 0)
		break;
	}
	if (i == ci->akickcount) {
	    if (ci->akickcount >= conf.csakicks) {
		notice(s_ChanServ, source,
		    "Sorry, you can only have %d autokicks on a channel.",
		   conf.csakicks);
		return;
	    }
	    ci->akickcount++;
	    if (debug >= 3) {
		log("debug: %s: cs_akick(add): reallocing akick list to %d (old list=%p)",
			s_ChanServ, ci->akickcount, ci->akick);
	    }
	    ci->akick = srealloc(ci->akick, sizeof(AutoKick) * ci->akickcount);
	    if (debug >= 3) {
		log("debug: %s: cs_akick(add): reallocing akick list to %d (new list %p)",
			s_ChanServ, ci->akickcount, ci->akick);
	    }
	}
	akick = &ci->akick[i];
	akick->name = (ni ? sstrdup(mask) : mask);
	akick->is_nick = (ni != NULL);
	akick->pad = 0;
	if (reason)
	    akick->reason = sstrdup(reason);
	else
	    akick->reason = NULL;
	if (debug >= 3) {
	    log("debug: %s: cs_akick(add): confirmation message (akick %d = %p, name %p, mask %p, ni %p)",
			s_ChanServ, i, akick, akick->name, mask, ni);
	}
	if (ci->flags & CI_VERBOSE) {
            send_cmd(NULL, ":%s NOTICE @%s :[VERBOSE] \2%s\2!%s@%s => akick %s %s %s", s_ChanServ, chan, source, u->username, u->host, chan, cmd, mask);
        }
	notice(s_ChanServ, source,
		"\2%s\2 added to %s autokick list.", akick->name, chan);

    } else if (stricmp(cmd, "DEL") == 0) {

	/* Special case: is it a valid number?  Only do search if it isn't. */
	if (strspn(mask, "1234567890") == strlen(mask) &&
				(i = atoi(mask)) > 0 && i <= ci->akickcount) {
	    i--;
	    if (ci->akick[i].is_nick < 0) {
            if (ci->flags & CI_VERBOSE) {
                send_cmd(NULL, ":%s NOTICE @%s :[VERBOSE] \2%s\2!%s@%s => akick %s %s %s", s_ChanServ, chan, source, u->username, u->host, chan, cmd, mask);
                }
		notice(s_ChanServ, source,
			"No such entry (#%d) on %s autokick list.", i+1, chan);
		return;
	    }
	} else {
	    /* First try for an exact match; then, a case-insensitive one. */
	    for (i = 0; i < ci->akickcount; i++) {
		if (ci->akick[i].is_nick < 0)
		    continue;
		if (strcmp(ci->akick[i].name, mask) == 0)
		    break;
	    }
	    if (i == ci->akickcount) {
		for (i = 0; i < ci->akickcount; i++) {
		    if (ci->akick[i].is_nick < 0)
			continue;
		    if (stricmp(ci->akick[i].name, mask) == 0)
			break;
		}
	    }
	    if (i == ci->akickcount) {
		notice(s_ChanServ, source,
			"\2%s\2 not found on %s autokick list.", mask, chan);
		return;
	    }
	}
	akick = &ci->akick[i];
	akick->is_nick = -1;
	notice(s_ChanServ, source,
		"\2%s\2 deleted from %s autokick list.", akick->name, chan);
	free(akick->name);
	akick->name = NULL;
	if (akick->reason) {
	    free(akick->reason);
	    akick->reason = NULL;
	}
	if (i == ci->akickcount-1) {
	    while (i >= 0 && ci->akick[i].in_use < 0)
		i--;
	    i++;
	    ci->akickcount = i;
	    if (i == 0) {
		free(ci->akick);
		ci->akick = NULL;
	    } else {
		ci->akick = srealloc(ci->akick, sizeof(AutoKick) * i);
	    }
	}

    } else if (stricmp(cmd, "LIST") == 0) {

	notice(s_ChanServ, source, "Autokick list for %s:", chan);
      notice(s_ChanServ, source, "Num Mask");
	for (akick = ci->akick, i = 0; i < ci->akickcount; akick++, i++) {
	    if (akick->is_nick < 0)
		continue;
	    if (mask && !match_wild(mask, akick->name))
		continue;
	    if (ni = findnick(akick->name))
		t = ni->last_usermask;
	    else
		t = NULL;
	    notice(s_ChanServ, source, "%3d %s%s%s%s%s%s%s",
			i+1, akick->name,
			t ? " (" : "", t ? t : "", t ? ")" : "",
			akick->reason ? " (" : "",
			akick->reason ? akick->reason : "",
			akick->reason ? ")" : "");
	}

    } else {

	notice(s_ChanServ, source,
		"Syntax: AKICK %s [ADD|DEL|LIST] <user@host> <\37nick\27>",
		*chan == '#' ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP AKICK\2.",
			s_ChanServ);

    }
}

/*************************************************************************/

static void cs_levels(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *cmd  = strtok(NULL, " ");
    char *what = strtok(NULL, " ");
    char *s    = strtok(NULL, " ");
    ChannelInfo *ci;
    User *u;
    short level;
    int i;

    /* If SET, we want two extra parameters; if DIS[ABLE], we want only
     * one; else, we want none.
     */
    if (!cmd || ((stricmp(cmd,"SET")==0) ? !s : (strnicmp(cmd,"DIS",3)==0) ? (!what || s) : !!what)) {
	notice(s_ChanServ, source, "Syntax: LEVELS %s [SET|DISABLE|LIST|RESET] <item> <level>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP LEVELS\2.", s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if ((!(u = finduser(source)) || !is_founder(u, ci)) &&
               !(stricmp(cmd, "LIST") == 0 && is_oper(source))) {
	notice(s_ChanServ, source, "Access denied.");
    } else if (stricmp(cmd, "SET") == 0) {
	level = atoi(s);
	if (level <= ACCESS_INVALID || level >= ACCESS_FOUNDER) {
	    notice(s_ChanServ, source, "Level must be between %d and %d inclusive.", ACCESS_INVALID+1, ACCESS_FOUNDER-1);
	    return;
	}
	for (i = 0; levelinfo[i].what >= 0; i++) {
	    if (levelinfo[i].settable == C_FALSE) {
		notice(s_ChanServ, source, "Level setting \2%s\2 on channel \2%s\2 is not settable (static).", levelinfo[i].name, chan);
		return;
	    }
	    if (stricmp(levelinfo[i].name, what) == 0) {
		ci->levels[levelinfo[i].what] = level;
		notice(s_ChanServ, source, "Level setting \2%s\2 on channel \2%s\2 changed to \2%d\2.", levelinfo[i].name, chan, level);
		return;
	    }
	}
	notice(s_ChanServ, source, "Level setting \2%s\2 is unknown, please type \2/msg %s HELP LEVELS DESC\2"
		" for a list of valid level settings.", what, s_ChanServ);
    } else if (stricmp(cmd, "DIS") == 0 || stricmp(cmd, "DISABLE") == 0) {
	for (i = 0; levelinfo[i].what >= 0; i++) {
	    if (levelinfo[i].settable == C_FALSE) {
		notice(s_ChanServ, source, "Level setting \2%s\2 on channel \2%s\2 is not settable (static).", levelinfo[i].name, chan);
		return;
	    }
	    if (stricmp(levelinfo[i].name, what) == 0) {
		ci->levels[levelinfo[i].what] = ACCESS_INVALID;
		notice(s_ChanServ, source, "Level setting \2%s\2 disabled on channel \2%s\2.", levelinfo[i].name, chan);
		return;
	    }
	}
	notice(s_ChanServ, source, "Level setting \2%s\2 is unknown, please yype \2/msg %s HELP LEVELS DESC\2"
		" for a list of valid level settings.", what, s_ChanServ);

    } else if (stricmp(cmd, "LIST") == 0) {

	int i;
	notice(s_ChanServ, source,
		"Access level settings for channel \2%s\2:", chan);
	if (!levelinfo_maxwidth) {
	    for (i = 0; levelinfo[i].what >= 0; i++) {
		int len = strlen(levelinfo[i].name);
		if (len > levelinfo_maxwidth)
		    levelinfo_maxwidth = len;
	    }
	}
	for (i = 0; levelinfo[i].what >= 0; i++) {
	    int j = ci->levels[levelinfo[i].what];
	    if (j == ACCESS_INVALID) {
		j = levelinfo[i].what;
		if (j == CA_AUTOOP || j == CA_AUTODEOP
				|| j == CA_AUTOVOICE || j == CA_NOJOIN)
		    s = "(disabled)";
		else
		    s = "(founder only)";
		notice(s_ChanServ, source, "    %-*s  %s",
			levelinfo_maxwidth, levelinfo[i].name, s);
	    } else {
		notice(s_ChanServ, source, "    %-*s  %d",
			levelinfo_maxwidth, levelinfo[i].name, j);
	    }
	}
    } else if (stricmp(cmd, "RESET") == 0) {
	reset_levels(ci);
	notice(s_ChanServ, source, "Access levels for \2%s\2 reset to defaults.", chan);
    } else {
	notice(s_ChanServ, source, "Syntax: LEVELS %s [SET|DISABLE|LIST|RESET] <item> <level>",
		(chan && *chan == '#') ? chan : "<channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP LEVELS\2.", s_ChanServ);
    }
}

/*************************************************************************/

static void cs_info(const char *source)
{
    char *chan = strtok(NULL, " ");
    ChannelInfo *ci;
    NickInfo *ni;
    struct tm tm;
    char s[BUFSIZE], *t, *end;
    Channel *c; 
    int ops, users, voices, halfops;
    struct c_userlist *data;

    if (!chan) {
	notice(s_ChanServ, source, "Syntax: INFO <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP INFO\2.", s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (ci->flags & CI_VERBOTEN) {
	notice(s_ChanServ, source, "Channel \2%s\2 may not be registered or used.", chan);
    } else {
	notice(s_ChanServ, source, "Information for channel \2%s\2:", chan);
	if (ni = findnick(ci->founder))
	    t = ni->last_usermask;
	else
	    t = NULL;
	notice(s_ChanServ, source, "        Founder: %s%s%s%s", ci->founder, t ? " (" : "", t ? t : "", t ? ")" : "");
      notice(s_ChanServ, source, "    Description: %s", ci->desc);
	tm = *localtime(&ci->time_registered);
	notice(s_ChanServ, source, "     Registered: %s %2d %02d:%02d:%02d %d", month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	tm = *localtime(&ci->last_used);
	notice(s_ChanServ, source, "      Last used: %s %2d %02d:%02d:%02d %d", month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	if (ci->last_topic) {
	notice(s_ChanServ, source, "     Last topic: %s", ci->last_topic);
      notice(s_ChanServ, source, "   Topic set by: %s\n", ci->last_topic_setter);
	}
	if (!(ci->flags & CI_PRIVATE)) {
		c = findchan (chan);
	    if (c) {
		for (data = c->users, users = 0; data; data = data->next)
		    users++;
		for (data = c->voices, voices = 0; data; data = data->next)
		    voices++;
		for (data = c->chanops, ops = 0; data; data = data->next)
		    ops++;
#ifdef V_UNREAL
		for (data = c->halfops, halfops = 0; data; data = data->next)
		    halfops++;
		notice (s_ChanServ, source, "     Statistics: %d op%s, %d halfop%s, %d voice%s, %d user%s",
			ops, ops == 1 ? "" : "s",
			halfops, halfops == 1 ? "" : "s",
			voices, voices == 1 ? "" : "s",
			users, users == 1 ? "" : "s");
#else
		notice (s_ChanServ, source, "     Statistics: %d op%s, %d voice%s, %d user%s",
			ops, ops == 1 ? "" : "s",
			voices, voices == 1 ? "" : "s",
			users, users == 1 ? "" : "s");
#endif
	    }
	}
	if (ci->url)
	    notice(s_ChanServ, source,
		"            URL: %s\n", ci->url);
	if (ci->email)
	    notice(s_ChanServ, source,
		" E-mail address: %s\n", ci->email);
	if (!ci->flags)
	    strcpy(s, "None\n");
	else {
	    int need_comma = 0;
	    static const char commastr[] = ", ";
	    end = s;
	    if (ci->flags & CI_IRCOP && !(ci->flags & CI_ADMIN)) {
		end += snprintf(end, sizeof(s)-(end-s), "IRC Operators");
		need_comma = 1;
	    }
	    if (ci->flags & CI_ADMIN) {
		end += snprintf(end, sizeof(s)-(end-s), "%sServices Administrators", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_EXPIRY) {
		end += snprintf(end, sizeof(s)-(end-s), "%sExpires", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_FROZEN) {
		end += snprintf(end, sizeof(s)-(end-s), "%sFrozen", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_CHELP) {
		end += snprintf(end, sizeof(s)-(end-s), "%sHelp Channel", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_MARK && is_oper(source)) {
		end += snprintf(end, sizeof(s)-(end-s), "%sMarked (%s)", need_comma ? commastr : "", ci->mark_nick);
		need_comma = 1;
	    }
	    if (ci->flags & CI_KEEPTOPIC) {
		end += snprintf(end, sizeof(s)-(end-s), "%sTopic Retention", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_TOPICLOCK) {
		end += snprintf(end, sizeof(s)-(end-s), "%sTopic Lock", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_SECUREOPS) {
		end += snprintf(end, sizeof(s)-(end-s), "%sSecure Ops", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_RESTRICTED) {
		end += snprintf(end, sizeof(s)-(end-s), "%sRestricted Access", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_SECURE) {
		end += snprintf(end, sizeof(s)-(end-s), "%sSecure", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_PRIVATE) {
		end += snprintf(end, sizeof(s)-(end-s), "%sPrivate", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_OPGUARD) {
		end += snprintf(end, sizeof(s)-(end-s), "%sOpGuard", need_comma ? commastr : "");
		need_comma = 1;
	    }
	    if (ci->flags & CI_VERBOSE) {
		end += snprintf(end, sizeof(s)-(end-s), "%sVerbose", need_comma ? commastr : "");
		need_comma = 1;
	    }
	}
	notice(s_ChanServ, source, "        Options: %s", s);
	end = s;
	*end = 0;
	if (ci->mlock_on || ci->mlock_key || ci->mlock_limit)
	    end += sprintf(end, "+%s%s%s%s%s%s%s%s%s%s%s",
#ifdef V_BAHAMUT
				(ci->mlock_on & CMODE_C) ? "c" : "",
#else
				"",
#endif
#ifdef V_UNREAL
				(ci->mlock_on & CMODE_H) ? "h" : "",
#else
				"",
#endif
				(ci->mlock_on & CMODE_I) ? "i" : "",
				(ci->mlock_key         ) ? "k" : "",
				(ci->mlock_limit       ) ? "l" : "",
				(ci->mlock_on & CMODE_M) ? "m" : "",
				(ci->mlock_on & CMODE_N) ? "n" : "",
				(ci->mlock_on & CMODE_P) ? "p" : "",
				(ci->mlock_on & CMODE_S) ? "s" : "",
				(ci->mlock_on & CMODE_R) ? "r" : "",
				(ci->mlock_on & CMODE_T) ? "t" : "");
	if (ci->mlock_off)
	    end += sprintf(end, "-%s%s%s%s%s%s%s%s%s%s%s",
#ifdef V_BAHAMUT
				(ci->mlock_off & CMODE_C) ? "c" : "",
#else
				"",
#endif
#ifdef V_UNREAL
				(ci->mlock_off & CMODE_H) ? "h" : "",
#else
				"",
#endif
				(ci->mlock_off & CMODE_I) ? "i" : "",
				(ci->mlock_off & CMODE_K) ? "k" : "",
				(ci->mlock_off & CMODE_L) ? "l" : "",
				(ci->mlock_off & CMODE_M) ? "m" : "",
				(ci->mlock_off & CMODE_N) ? "n" : "",
				(ci->mlock_off & CMODE_P) ? "p" : "",
				(ci->mlock_off & CMODE_S) ? "s" : "",
				(ci->mlock_off & CMODE_R) ? "r" : "",
				(ci->mlock_off & CMODE_T) ? "t" : "");
	if (s)
	    notice(s_ChanServ, source, "      Mode lock: %s", s);
    }
}

/*************************************************************************/

static void cs_list(const char *source)
{
    char *pattern = strtok(NULL, " ");
    ChannelInfo *ci;
    int nchans, i;
    char buf[BUFSIZE];
    int chancnt = 0;

    if (conf.nslistoper) {
       User *u = finduser(source);
       if (!u || !(u->mode & UMODE_O)) {
	    notice(s_NickServ, source, "Access denied.");
	    return;
       }
    }
    if (!pattern) {

	notice(s_ChanServ, source, "Syntax: LIST <pattern>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP LIST\2.",
		s_ChanServ);
    } else {
	nchans = 0;
	notice(s_ChanServ, source, "List of entries matching \2%s\2:", pattern);
      notice(s_ChanServ, source, "Num Channel");
	for (i = 0; i < 256; i++) {
	    for (ci = chanlists[i]; ci; ci = ci->next) {
		/* We hide private, forbidden, closed and protected channels. */
		if (ci->flags & (CI_PRIVATE | CI_VERBOTEN | CI_CLOSE | CI_PROTECTED))
		    continue;
 		chancnt++;
		snprintf(buf, sizeof(buf), "%3d %s (%s)", chancnt, ci->name, ci->desc);
		if (stricmp(pattern, ci->name) == 0 ||
					match_wild_nocase(pattern, buf)) {
		    if (++nchans <= 50)
			notice(s_ChanServ, source, "%s", buf);
		}
	    }
	}
	notice(s_ChanServ, source, "End of list - %d/%d matches shown.",
		nchans>50 ? 50 : nchans, nchans);
    }
}

/*************************************************************************/

static void cs_invite(const char *source)
{
    char *chan = strtok(NULL, " ");
    User *u = finduser(source);
    Channel *c;
    ChannelInfo *ci;

    if (!chan) {
	notice(s_ChanServ, source, "Syntax: INVITE <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP INVITE\2.",
			s_ChanServ);
    } else if (!(c = findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 does not exist.", chan);
    } else if (!(ci = c->ci)) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u || !check_access(u, ci, CA_INVITE)) {
	notice(s_ChanServ, source, "Access denied.");
    } else {
	send_cmd(s_ChanServ, "INVITE %s %s", source, chan);
    }
}

/*************************************************************************/

static void cs_op(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *op_params = strtok(NULL, " ");  
    User *u = finduser(source);
    ChannelInfo *ci;

    if (!chan || !op_params) {
	notice(s_ChanServ, source, "Syntax: OP <channel> <nick>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP OP\2.", s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u || !check_access(u, ci, CA_OPDEOP)) {
	notice(s_ChanServ, source, "Access denied.");
    } else {
      if (is_on_chan(op_params, chan)) {
		char *av[3];
		send_cmd(s_ChanServ, "MODE %s +o %s", chan, op_params);
		av[0] = chan;
		av[1] = sstrdup("+o");
		av[2] = op_params;
		do_cmode(s_ChanServ, 3, av);
		free(av[1]);
		if (stricmp(op_params, source) != 0)
			send_cmd(s_ChanServ, "NOTICE %s :%s!%s@%s has opped %s.", chan, source, u->username, u->host, op_params);
	} else {
	      notice(s_ChanServ, source, "Nick \2%s\2 is not currently in \2%s\2.", op_params, chan);
	}
   }
}

/*************************************************************************/

static void cs_voice(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *voice_params = strtok(NULL, " ");  
    User *u = finduser(source);
    ChannelInfo *ci;

    if (!chan || !voice_params) {
	notice(s_ChanServ, source, "Syntax: VOICE <channel> <nick>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP VOICE\2.", s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u || !check_access(u, ci, CA_AUTOVOICE)) {
	notice(s_ChanServ, source, "Access denied.");
    } else {
      if (is_on_chan(voice_params, chan)) {
		char *av[3];
		send_cmd(s_ChanServ, "MODE %s +v %s", chan, voice_params);
		av[0] = chan;
		av[1] = sstrdup("+o");
		av[2] = voice_params;
		do_cmode(s_ChanServ, 3, av);
		free(av[1]);
		if (stricmp(voice_params, source) != 0)
			send_cmd(s_ChanServ, "NOTICE %s :%s!%s@%s has voiced %s.", chan, source, u->username, u->host, voice_params);
	} else {
	      notice(s_ChanServ, source, "Nick \2%s\2 is not currently in \2%s\2.", voice_params, chan);
	}
   }
}

/*************************************************************************/

static void cs_devoice(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *devoice_params = strtok(NULL, " ");  
    User *u = finduser(source);
    ChannelInfo *ci;

    if (!chan || !devoice_params) {
	notice(s_ChanServ, source, "Syntax: DEVOICE <channel> <nick>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP DEVOICE\2.", s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u || !check_access(u, ci, CA_AUTOVOICE)) {
	notice(s_ChanServ, source, "Access denied.");
    } else {
      if (is_on_chan(devoice_params, chan)) {
		char *av[3];
		send_cmd(s_ChanServ, "MODE %s -v %s", chan, devoice_params);
		av[0] = chan;
		av[1] = sstrdup("+o");
		av[2] = devoice_params;
		do_cmode(s_ChanServ, 3, av);
		free(av[1]);
		if (stricmp(devoice_params, source) != 0)
			send_cmd(s_ChanServ, "NOTICE %s :%s!%s@%s has devoiced %s.", chan, source, u->username, u->host, devoice_params);
	} else {
	      notice(s_ChanServ, source, "Nick \2%s\2 is not currently in \2%s\2.", devoice_params, chan);
	}
   }
}
/*************************************************************************/

#ifdef V_UNREAL
static void cs_halfop(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *halfop_params = strtok(NULL, " ");  
    User *u = finduser(source);
    ChannelInfo *ci;

    if (!chan || !halfop_params) {
	notice(s_ChanServ, source, "Syntax: HALFOP <channel> <nick>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP HALFOP\2.", s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u || !check_access(u, ci, CA_AUTOHALFOP)) {
	notice(s_ChanServ, source, "Access denied.");
    } else {
      if (is_on_chan(halfop_params, chan)) {
		char *av[3];
		send_cmd(s_ChanServ, "MODE %s +h %s", chan, halfop_params);
		av[0] = chan;
		av[1] = sstrdup("+h");
		av[2] = halfop_params;
		do_cmode(s_ChanServ, 3, av);
		free(av[1]);
		if (stricmp(halfop_params, source) != 0)
			send_cmd(s_ChanServ, "NOTICE %s :%s!%s@%s has half opped %s.", chan, source, u->username, u->host, halfop_params);
	} else {
	      notice(s_ChanServ, source, "Nick \2%s\2 is not currently in \2%s\2.", halfop_params, chan);
	}
   }
}

/*************************************************************************/

static void cs_dehalfop(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *dehalfop_params = strtok(NULL, " ");
    User *u = finduser(source);
    ChannelInfo *ci;

    if (!chan || !dehalfop_params) {
	notice(s_ChanServ, source, "Syntax: DEHALFOP <channel> <nick>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP DEHALFOP\2.",
		s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u || !check_access(u, ci, CA_AUTOHALFOP)) {
	notice(s_ChanServ, source, "Access denied.");
    } else {
      if (is_on_chan(dehalfop_params, chan)) {
		char *av[3];
		send_cmd(s_ChanServ, "MODE %s -h %s", chan, dehalfop_params);
		av[0] = chan;
		av[1] = sstrdup("-h");
		av[2] = dehalfop_params;
		do_cmode(s_ChanServ, 3, av);
		free(av[1]);
		if (stricmp(dehalfop_params, source) != 0)
			send_cmd(s_ChanServ, "NOTICE %s :%s!%s@%s has dehalfopped %s.", chan, source, u->username, u->host, dehalfop_params);
	} else {
	      notice(s_ChanServ, source, "Nick \2%s\2 is not currently in \2%s\2.", dehalfop_params, chan);
	}
   }
}

#endif

/*************************************************************************/

static void cs_deop(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *deop_params = strtok(NULL, " ");
    User *u = finduser(source);
    ChannelInfo *ci;

    if (!chan || !deop_params) {
	notice(s_ChanServ, source, "Syntax: DEOP <channel> <nick>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP DEOP\2.",
		s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u || !check_access(u, ci, CA_OPDEOP)) {
	notice(s_ChanServ, source, "Access denied.");
    } else {
      if (is_on_chan(deop_params, chan)) {
		char *av[3];
		send_cmd(s_ChanServ, "MODE %s -o %s", chan, deop_params);
		av[0] = chan;
		av[1] = sstrdup("-o");
		av[2] = deop_params;
		do_cmode(s_ChanServ, 3, av);
		free(av[1]);
		if (stricmp(deop_params, source) != 0)
			send_cmd(s_ChanServ, "NOTICE %s :%s!%s@%s has deopped %s.", chan, source, u->username, u->host, deop_params);
	} else {
	      notice(s_ChanServ, source, "Nick \2%s\2 is not currently in \2%s\2.", deop_params, chan);
	}
   }
}

/*************************************************************************/

static void cs_unban(const char *source)
{
    char *chan = strtok(NULL, " ");
    User *u = finduser(source);
    ChannelInfo *ci;
    Channel *c;
    int i;
    char *av[3];

    if (!chan) {
	notice(s_ChanServ, source, "Syntax: UNBAN <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP UNBAN\2.",
		s_ChanServ);
    } else if (!(c = findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 does not exist.", chan);
    } else if (!(ci = c->ci)) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u || !check_access(u, ci, CA_UNBAN)) {
	notice(s_ChanServ, source, "Access denied.");
    } if (ci->flags & CI_VERBOSE) {
     send_cmd(NULL, ":%s NOTICE @%s :[VERBOSE] \2%s\2!%s@%s => unban", s_ChanServ, chan, source, u->username, u->host);
    } else {
	int count = c->bancount;
	char **bans = smalloc(sizeof(char *) * count);
	memcpy(bans, c->bans, sizeof(char *) * count);

	av[0] = chan;
	av[1] = sstrdup("-b");
	for (i = 0; i < count; i++) {
	    if (match_usermask(bans[i], u)) {
		send_cmd(s_ChanServ, "MODE %s -b %s", chan, bans[i]);
		av[2] = sstrdup(bans[i]);
		do_cmode(s_ChanServ, 3, av);
		free(av[2]);
	    }
	}
	free(av[1]);
	notice(s_ChanServ, source, "You have been unbanned from the channel \2%s\2.", chan);
	free(bans);
    }
}

/*************************************************************************/

/* cs_mdeop()
 * Massdeop all users in a channel. */

static void cs_mdeop(const char *source)
{
   char *chan = strtok(NULL, " ");
   ChannelInfo *ci;
   User *u;
   Channel *c;
   Timeout *t;
   char *av[3];
   struct c_userlist *cu, *next;
  
   if (!chan) {
      notice(s_ChanServ, source, "Syntax: MDEOP <channel>");
      notice(s_ChanServ, source, "For more information type \2/msg %s HELP MDEOP.", s_ChanServ);
   } else if (!(ci = cs_findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
   } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
   } else if (!(c = findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 does not exist.", chan);
   } else if (!check_access(u, ci, CA_AKICK)) {
	notice(s_ChanServ, source, "Access denied.");
   } else {
      for (cu = c->chanops; cu; cu = next) {
         next = cu->next;
         av[0] = sstrdup(chan);
         av[1] = sstrdup("-o");
         av[2] = sstrdup(cu->user->nick);
         send_cmd(s_ChanServ, "MODE %s %s :%s", av[0], av[1], av[2]);
         do_cmode(s_ChanServ, 3, av);
         free(av[2]);
         free(av[1]);
         free(av[0]);
      }
      send_cmd(s_ChanServ, "JOIN %s", chan);
      send_cmd(s_ChanServ, "MODE %s +o %s", chan, s_ChanServ);
      t = add_timeout(conf.csinhabit, timeout_leave, 0);
      t->data = sstrdup(chan);
      notice(s_ChanServ, source, "Channel \2%s\2 has been massdeoped.", chan);
   }
}

/*************************************************************************/

/* cs_mkick()
 * Masskick all users in a channel. */

static void cs_mkick(const char *source)
{
   char *chan = strtok(NULL, " ");
   ChannelInfo *ci;
   User *u;
   Channel *c;
   Timeout *t;
   char *av[3];
   struct c_userlist *cu, *next;
  
   if (!chan) {
      notice(s_ChanServ, source, "Syntax: MDEOP <channel>");
      notice(s_ChanServ, source, "For more information type \2/msg %s HELP MDEOP.", s_ChanServ);
   } else if (!(ci = cs_findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
   } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
   } else if (!(c = findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 does not exist.", chan);
   } else if (!check_access(u, ci, CA_AKICK)) {
	notice(s_ChanServ, source, "Access denied.");
   } else {
      for (cu = c->chanops; cu; cu = next) {
         next = cu->next;
         av[0] = sstrdup(chan);
         av[1] = sstrdup("-o");
         av[2] = sstrdup(cu->user->nick);
         send_cmd(s_ChanServ, "KICK %s %s :%s", av[0], av[1], av[2]);
         do_kick(s_ChanServ, 3, av);
         free(av[2]);
         free(av[1]);
         free(av[0]);
      }
      send_cmd(s_ChanServ, "JOIN %s", chan);
      send_cmd(s_ChanServ, "MODE %s +o %s", chan, s_ChanServ);
      send_cmd(s_ChanServ, "MODE %s +isb *!*@*", chan);
      t = add_timeout(conf.csinhabit, timeout_leave, 0);
      t->data = sstrdup(chan);
      notice(s_ChanServ, source, "Channel \2%s\2 has been masskicked.", chan);
   }
}

/*************************************************************************/

static void cs_getpass(const char *source)
{
    char *chan = strtok(NULL, " ");
    User *u = finduser(source);
    ChannelInfo *ci;

    if (!chan) {
	notice(s_ChanServ, source, "Syntax: GETPASS <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP GETPASS\2.",
			s_ChanServ);
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!u) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
    } else {
	log("%s: %s!%s@%s retrieved (getpass) the password for the channel %s.",
	   s_ChanServ, source, u->username, u->host, chan);
	globops(s_ChanServ, "\2%s\2 retrieved (getpass) the password for the channel \2%s\2",
		source, chan);
	notice(s_ChanServ, source, "Password for the channel \2%s\2 is \2%s\2",
		chan, ci->founderpass);
    }
}

/*************************************************************************/

static void cs_forbid(const char *source)
{
    ChannelInfo *ci;
    char *chan = strtok(NULL, " ");
    User *u;

    if (!chan) {
	notice(s_ChanServ, source, "Syntax: FORBID <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP FORBID\2.", s_ChanServ);
	return;
    } else if (!(u = finduser(source))) {
      notice(s_NickServ, source, "Could not retrieve your online user information.");
    } else {
       if (ci = cs_findchan(chan))
	    delchan(ci);
       if (ci = makechan(chan)) {
		log("%s: %s!%s@%s has forbidden (forbid) the channel %s.", s_ChanServ, 
			source, u->username, u->host, chan);
		ci->flags |= CI_VERBOTEN;
      	globops(s_ChanServ, "\2%s\2 has forbidden (forbid) the channel \2%s\2.", source, chan);
		notice(s_ChanServ, source, "Channel \2%s\2 has been forbidden.", chan);
	 } else {
		log("%s: makechan() failed for the forbidding of the channel %s.", s_ChanServ, chan);
		notice(s_ChanServ, source, "Could not forbid the channel \2%s\2.", chan);
	 }
    }
}

/*************************************************************************/

static void cs_status(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *nick = strtok(NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!nick || !chan) {
	notice(s_ChanServ, source, "Syntax: STATUS <channel> <nick>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP STATUS\2.", s_ChanServ);
    } else if (!(u = finduser(source))) {
	notice(s_ChanServ, source, "Could not retrieve your online user information.");
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (!(u = finduser(nick))) {
	notice(s_ChanServ, source, "Nick \2%s\2 is not currently online", nick);
    } else {
      notice(s_ChanServ, source, "The status of \2%s\2 on the channel \2%s\2 is \2%d\2.", nick, chan, get_access(u, ci));
    }
}

/*************************************************************************/

static void cs_expiry(const char *source)
{
    char *param = strtok(NULL, " ");
    char *chan = strtok(NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!chan || !param) {
	notice(s_ChanServ, source, "Syntax: EXPIRY [ENABLE|DISABLE] <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP EXPIRY\2.", s_ChanServ);
    } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (stricmp(param, "ENABLE") == 0) {
	ci->flags |= CI_EXPIRY;
      log("%s: %s!%s@%s has enabled expiry (expiry) on the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has enabled expiry (expiry) on the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Expiry on the channel \2%s\2 has been enabled.", chan);
    } else if (stricmp(param, "DISABLE") == 0) {
	ci->flags &= ~CI_EXPIRY;
      log("%s: %s!%s@%s has disabled expiry (expiry) for the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has disabled expiry (expiry) for the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Expiry for the channel \2%s\2 has been disabled.", chan);
    } else {
	notice(s_ChanServ, source, "Syntax: EXPIRY [ENABLE|DISABLE] <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP EXPIRY\2.", s_ChanServ);
    }
}

/*************************************************************************/

/* cs_mark()
 * Mark the a channel so only the CSOp who marked it may deal with issues of that channel. */

static void cs_mark(const char *source)
{
    char *param = strtok(NULL, " ");
    char *chan= strtok(NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!chan || !param) {
	notice(s_ChanServ, source, "Syntax: MARK [ENABLE|DISABLE] <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP MARK\2.", s_ChanServ);
    } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (stricmp(param, "ENABLE") == 0) {
	ci->flags |= CI_MARK;
	strscpy(ci->mark_nick, u->nick, NICKMAX);
      log("%s: %s!%s@%s has enabled a mark (mark) on the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has enabled a mark (mark) on the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Mark on the channel \2%s\2 has been enabled.", chan);
    } else if (stricmp(param, "DISABLE") == 0) {
	ci->flags &= ~CI_MARK;
      log("%s: %s!%s@%s has disabled the mark (mark) for the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has disabled the mark (mark) for the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Mark for the channel \2%s\2 has been disabled.", chan);
    } else {
	notice(s_ChanServ, source, "Syntax: MARK [ENABLE|DISABLE] <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP MARK\2.", s_ChanServ);
    }
}

/*************************************************************************/

/* cs_freeze()
 * Freeze a channel, prevent users on the access or channel op lists from using their priviledges. */

static void cs_freeze(const char *source)
{
    char *param = strtok(NULL, " ");
    char *chan= strtok(NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!chan || !param) {
	notice(s_ChanServ, source, "Syntax: FREEZE [ENABLE|DISABLE] <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP FREEZE\2.", s_ChanServ);
    }else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (stricmp(param, "ENABLE") == 0) {
	ci->flags |= CI_FROZEN;
      log("%s: %s!%s@%s has enabled a freeze (freeze) on the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has enabled a freeze (freeze) on the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Freeze on the channel \2%s\2 has been enabled.", chan);
    } else if (stricmp(param, "DISABLE") == 0) {
	ci->flags &= ~CI_FROZEN;
      log("%s: %s!%s@%s has disabled the freeze (freeze) for the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has disabled the freeze (freeze) for the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Freeze for the channel \2%s\2 has been disabled.", chan);
    } else {
	notice(s_ChanServ, source, "Syntax: FREEZE [ENABLE|DISABLE] <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP FREEZE\2.", s_ChanServ);
    }
}

/*************************************************************************/

/* cs_setpass()
 * Set the password for a channel. */

static void cs_setpass(const char *source)
{
   char *chan = strtok(NULL, " ");
   char *pass = strtok(NULL, " ");
   ChannelInfo *ci;
   User *u;

   if (!chan || !pass) {
	notice(s_ChanServ, source, "Syntax: SETPASS <channel> <password>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP SETPASS\2.", s_ChanServ);
   } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
   } else if (!(ci = cs_findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
   } else {
      log("%s: %s!%s@%s has changed (setpass) the password for the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      strscpy(ci->founderpass, pass, PASSMAX);
      globops(s_ChanServ, "\2%s\2 has changed (setpass) the password for the channel \2%s\2.", source, chan);
      notice(s_ChanServ, source, "Password for the channel \2%s\2 changed to \2%s\2.", chan, pass);
   }
}

/*************************************************************************/

/* cs_getkey()
 * Get the key for a channel. */

static void cs_getkey(const char *source)
{
   char *chan = strtok(NULL, " ");
   Channel *c;
   User *u;

   if (!chan) {
	notice(s_ChanServ, source, "Syntax: GETKEY <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP GETKEY\2.", s_ChanServ);
   } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
   } else if (!(c = findchan(chan))) {
      notice(s_ChanServ, source, "Channel \2%s\2 does not exist.", chan);
   } else {
      log("%s: %s!%s@%s retrieved the key (getkey) for the channel %s.", s_ChanServ, source, u->username, u->host, chan);
	if (c->key != NULL) {
		globops(s_ChanServ, "\2%s\2 retrieved (getkey) the key for the channel \2%s\2",
			source, chan);
	      notice(s_ChanServ, source, "Key for the channel \2%s\2 is \2%s\2.", chan, c->key);
	} else {
	      notice(s_ChanServ, source, "Key for the channel \2%s\2 does not exist.", chan);
	}
   }
}

/*************************************************************************/

/* cs_chelp()
 * Set channel as a network help channel. */

static void cs_chelp(const char *source)
{
    char *param = strtok(NULL, " ");
    char *chan= strtok(NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!chan || !param) {
	notice(s_ChanServ, source, "Syntax: CHELP [ENABLE|DISABLE] <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP CHELP\2.", s_ChanServ);
    } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (stricmp(param, "ENABLE") == 0) {
	ci->flags |= CI_CHELP;
      log("%s: %s!%s@%s has enabled help channel identity (chelp) on the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has enabled help channel identity (chelp) on the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Help channel identity on the channel \2%s\2 has been enabled.", chan);
    } else if (stricmp(param, "DISABLE") == 0) {
	ci->flags &= ~CI_CHELP;
      log("%s: %s!%s@%s has disabled help channel identity (chelp) for the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has disabled help channel identity (chelp) for the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Help channel identity  for the channel \2%s\2 has been disabled.", chan);
    } else {
	notice(s_ChanServ, source, "Syntax: CHELP [ENABLE|DISABLE] <channel>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP CHELP\2.", s_ChanServ);
    }
}

/*************************************************************************/

/* cs_close()
 * Close a channel, kinda like forbid except it is closed and shows up in the info. */

static void cs_close(const char *source)
{
    char *param = strtok(NULL, " ");
    char *chan = strtok(NULL, " ");
    char *topic = strtok(NULL, "");
    ChannelInfo *ci;
    User *u;

    if (!chan || !param) {
	notice(s_ChanServ, source, "Syntax: CLOSE [ENABLE|DISABLE] <channel> <\37topic\37>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP CLOSE\2.", s_ChanServ);
    } else if (!(u = finduser(source))) {
      notice(s_ChanServ, source, "Could not retrieve your online user information.");
    } else if (!(ci = cs_findchan(chan))) {
	notice(s_ChanServ, source, "Channel \2%s\2 is not registered.", chan);
    } else if (stricmp(param, "ENABLE") == 0) {
	ci->flags |= CI_CLOSE;
	ci->flags &= ~CI_EXPIRY;
	if (ci->desc)
	   free(ci->desc);
      ci->desc = sstrdup("\2This channel has been closed by Ozyx Administration.\2");
	if (ci->last_topic)
	   free(ci->last_topic);
      if (topic) {
		ci->last_topic = sstrdup(topic);
	      strscpy(ci->last_topic_setter, u->nick, NICKMAX);
      } else {
	      ci->last_topic = sstrdup("\2This channel has been closed by Ozyx Administration.\2");
	      strscpy(ci->last_topic_setter, s_ChanServ, NICKMAX);
	}
      log("%s: %s!%s@%s has enabled closure (close) on the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has enabled closure (close) on the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Closure for the channel \2%s\2 has been enabled.", chan);
    } else if (stricmp(param, "DISABLE") == 0) {
	ci->flags &= ~CI_CLOSE;
	ci->flags |= CI_EXPIRY;
      log("%s: %s!%s@%s has disabled closure (close) for the channel %s.", s_ChanServ, source, u->username, u->host, chan);
      globops(s_ChanServ, "\2%s\2 has disabled closure (close) for the channel \2%s\2.", source, chan);
	notice(s_ChanServ, source, "Closure for the channel \2%s\2 has been disabled.", chan);
    } else {
	notice(s_ChanServ, source, "Syntax: CLOSE [ENABLE|DISABLE] <channel> <\37topic\37>");
	notice(s_ChanServ, source, "For more information please type \2/msg %s HELP CLOSE\2.", s_ChanServ);
    }
}

/*************************************************************************/

#endif

/*************************************************************************/
