/* HelpServ functions.
 * src/helpserv.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to 
 * this software. */

/*************************************************************************/

#ifndef HELPSERV_C
#define HELPSERV_C

/*************************************************************************/

#include "services.h"
#include "version.h"
#include <sys/stat.h>

/*************************************************************************/

static void do_help(const char *whoami, const char *source, char *topic);

/*************************************************************************/

/* helpserv:  Main HelpServ routine.  `whoami' is what nick we should send
 * messages as: this won't necessarily be s_HelpServ, because other
 * routines call this one to display help files. */

void helpserv(const char *whoami, const char *source, char *buf)
{
   char *cmd, *topic, *s;

   topic = buf ? sstrdup(buf) : NULL;
   cmd = strtok(buf, " ");
   if (cmd && stricmp(cmd, "\1PING") == 0) {
      if (!(s = strtok(NULL, "")))
         s = "\1";
      notice(s_HelpServ, source, "\1PING %s", s);
   } else if (stricmp(cmd, "CREDITS") == 0) {
      notice(s_HelpServ, source, "Formatting help page, please wait...");
      notice(s_HelpServ, source, "\2***** HelpServ Help *****\2");
      notice(s_HelpServ, source, "%s %s", __ABRA_NAME__, __ABRA_VERSION__);
      notice(s_HelpServ, source, "%s", __ABRA_COPYRIGHT__);
      notice(s_HelpServ, source, "");
      notice(s_HelpServ, source, "Code written by: hsals (Brian@AlphaNine.com)");
      notice(s_HelpServ, source, "Note to talk to us on IRC, please message us, our nick is the same as the first part of our"
		" email address, ie: Brian@AlphaNine.net would result in a nick being hsals or join our company channel #darksiderz.");
      notice(s_HelpServ, source, "\2***** End Of Help *****\2");
   } else if (stricmp(cmd, "COPYRIGHT") == 0) {
      notice(s_HelpServ, source, "Formatting help page, please wait...");
      notice(s_HelpServ, source, "\2***** HelpServ Help *****\2");
      notice(s_HelpServ, source, "%s %s", __ABRA_NAME__, __ABRA_VERSION__);
      notice(s_HelpServ, source, "%s", __ABRA_COPYRIGHT__);
      notice(s_HelpServ, source, "\2***** End Of Help *****\2");
   } else {
      do_help(whoami, source, topic);
   }
   if (topic)
      free(topic);
}

/*************************************************************************/

/* do_help()
 * Return a help message requested by a user. */

static void do_help(const char *whoami, const char *source, char *topic)
{
   FILE *f;
   struct stat st;
   char buf[256], *ptr, *s;
   char *old_topic;
   NickInfo *ni;

   if (!topic || !*topic)
      topic = "help";
   old_topic = sstrdup(topic);

   /* As we copy path parts, (1) lowercase everything and (2) make sure
    * we don't let any special characters through -- this includes '.'
    * (which could get parent dir) or '/' (which couldn't _really_ do
    * anything nasty if we keep '.' out, but better to be on the safe
    * side).  Special characters turn into '_'. */

   if ((ni = findnick(source))) {
	if (ni->flags & NI_LANG_FRENCH) {
	   strscpy(buf, LANG_FRENCH_HS_DIR, sizeof(buf));
	} else {
	   strscpy(buf, HELPSERV_DIR, sizeof(buf));
	}
   } else {
	   strscpy(buf, HELPSERV_DIR, sizeof(buf));
   }

   /* Second part for destination, we need to locate the language of the help files
    * so we put a small timer on it. */

   ptr = buf + strlen(buf);
   for (s = strtok(topic, " "); s && ptr-buf < sizeof(buf)-1; s = strtok(NULL, " ")) {
      *ptr++ = '/';
      while (*s && ptr-buf < sizeof(buf)-1) {
         if (*s == '.' || *s == '/')
            *ptr++ = '_';
         else
            *ptr++ = tolower(*s);
         ++s;
      }
      *ptr = 0;
   }

   /* If we end up at a directory, go for an "index" file/dir if
    * possible. */
   while (ptr - buf < sizeof(buf) - 1 && stat(buf, &st) == 0 && S_ISDIR(st.st_mode)) {
      *ptr++ = '/';
      strscpy(ptr, "index", sizeof(buf) - (ptr-buf));
      ptr += strlen(ptr);
   }
   /* Send the file, if it exists. */
   if (!(f = fopen(buf, "r"))) {
      if (debug)
         log_perror("debug: Cannot open help file %s", buf);
      notice(whoami, source, "Sorry, no help available for \2%s\2.", old_topic);
      free(old_topic);
      return;
   }
   notice(whoami, source, "Formatting help page, please wait...");
   while (fgets(buf, sizeof(buf), f)) {
      s = strtok(buf, "\n");
      /* Use this odd construction to prevent any %'s in the text from
       * doing weird stuff to the output.  Also replace blank lines by
       * spaces (see send.c/notice_list() for an explanation of why). */
      notice(whoami, source, "%s", s ? s : " ");
   }
   fclose(f);
   free(old_topic);
}

/*************************************************************************/

#endif

/*************************************************************************/
