/* Incoming IRC message handling functions and message lists.
 * src/messages.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to this software. */

/*************************************************************************/

#ifndef   __messages_source__
#define __messages_source__

#ifndef lint
   static char sccsid[] = "@(#)messages.c		7.0 18/03/01 (C) 2000-2001 DarkSiderz Net $exp 0";
#endif

/*************************************************************************/

#include "services.h"
#include "messages.h"

/*************************************************************************/

/* m_nickcoll()
 * Nick collisions, if a nick is one of DS's psuedo clients, restore it. */

static void m_nickcoll(char *source, int ac, char **av)
{
   if (ac < 1) {
      return;
   }
   if (!(run.flags & RUN_READONLY)) {
      introduce_user(av[0]);
   }
}

/*************************************************************************/

/* m_ping()
 * Send ping to server. */

static void m_ping(char *source, int ac, char **av)
{
   if (ac < 1) {
      return;
   }
   send_cmd(NULL, "PONG %s %s", ac > 1 ? av[1] : conf.lserver, av[0]);
}

/*************************************************************************/

/* m_pong()
 * Pong from server for lag check. */

static void m_pong(char *source, int ac, char **av)
{
   ServerInfo *si;

   if (ac < 1) {
      return;
   }
   if (ac > 1) {
      if (si = findserver(av[0])) {
         if (si->flags & SI_CONNECT && !(si->flags & SI_JUPE)) {
            si->ping = (time(NULL) - ping.last_sent);
	      /* The hub we connected to naturally is only 1 hop away, and we are really only concerned about this 
             * server's lag. */
	      if (si->hops == 1) {
	         if (si->ping >= 20 && si->ping < 45) {
                  globops(NULL, "DS is lagged %d seconds to its uplink %s, may consider rerouting.",
                     si->ping, si->server);
               }
               if (si->ping >= 45) {
                  globops(NULL, "DS is lagged %d seconds to its uplink %s, reroute is crucial.",
                      si->ping, si->server);
               }
	      }
         }
      }
   }
}

/*************************************************************************/

/* m_away()
 * Is user away or not? */

static void m_away(char *source, int ac, char **av)
{
   User *u;

   if (ac == 0 || *av[0] == 0) {
      if (u = finduser(source)) {
         check_memos(source);
         u->flags &= ~U_AWAY;
      }
   } else {
      if (u = finduser(source)) {
         u->flags = U_AWAY;
      }
   }
}

/*************************************************************************/

/* m_join()
 * User joining a channel. */

static void m_join(char *source, int ac, char **av)
{
   if (ac != 1) {
      return;
   }
   do_join(source, ac, av);
}

/*************************************************************************/

#ifdef V_BAHAMUT

/* m_sjoin()
 * User joining a channel. */

static void m_sjoin(char *source, int ac, char **av)
{
   do_sjoin(source, ac, av);
}

/*************************************************************************/

/* m_svinfo()
 * Server info. */

static void m_svinfo(char *source, int ac, char **av)
{
   do_svinfo(source, ac, av);
}

#endif

/*************************************************************************/

/* m_kick()
 * User being kicked from a channel. */

static void m_kick(char *source, int ac, char **av)
{
   if (ac != 3) {
      return;
   }
   do_kick(source, ac, av);
}

/*************************************************************************/

/* m_kill()
 * Nick being killed, if a nick is one of DS's psuedo clients, restore it. */

static void m_kill(char *source, int ac, char **av)
{
   if (ac != 2) {
      return;
   }
   if (is_services_client(av[0])) {
      introduce_user(av[0]);
   } else if (is_client(av[0])) {
      recover_client(av[0]);
   } else {
      do_kill(source, ac, av);
   }
}

/*************************************************************************/

/* m_mode()
 * User or channel changing modes. */

static void m_mode(char *source, int ac, char **av)
{
   /* Mode is being set for a channel, use do_cmode, otherwise for a user do_umode. */
   if (*av[0] == '#' || *av[0] == '&') {
      if (ac < 2) {
         return;
      }
      do_cmode(source, ac, av);
   } else {
      if (ac != 2) {
         return;
      }
      do_umode(source, ac, av);
   }
}

/*************************************************************************/

/* m_motd()
 * User requesting message of the day file. */

static void m_motd(char *source, int ac, char **av)
{
   FILE *f =  fopen(conf.motdfile, "r");
   char buf[BUFSIZE];

   send_cmd(NULL, "375 %s :- %s Message of the Day", source, conf.lserver);
   if (f) {
      while (fgets(buf, sizeof(buf), f)) {
         buf[strlen(buf) - 1] = 0;
         send_cmd(NULL, "372 %s :- %s", source, buf);
      }
      fclose(f);
   } else {
      send_cmd(NULL, "372 %s :- MOTD file is missing", source);
   }
   send_cmd(NULL, "376 %s :End of /MOTD command.", source);
}

/*************************************************************************/

static void m_nick(char *source, int ac, char **av)
{
#if defined(V_UNDERNET)
   if (strchr(source, '.')) {
      *source = 0;
   }
#elif defined(V_DF)
   if (ac == 8) {
	av[6] = av[7];
	ac--;
   }
   if ((!*source && ac != 7) || (*source && ac != 2)) {
	return;
   }
#endif
    do_nick(source, ac, av);
}

/*************************************************************************/

/* m_part()
 * User parting a channel. */

static void m_part(char *source, int ac, char **av)
{
   if (ac < 1 || ac > 2) {
      return;
   }
   do_part(source, ac, av);
}

/*************************************************************************/

#ifdef V_UNREAL

/* m_sethost()
 * Unreal setting the host of a particular user. */

static void m_sethost(char *source, int ac, char **av)
{
   User *u = finduser(source);

   if (u) {
	u->host = sstrdup(av[1]);
   }
}

#endif

/*************************************************************************/

/* m_privmsg()
 * User private messaging services. */

static void m_privmsg(char *source, int ac, char **av)
{
   User *u = finduser(source);

   if (ac != 2) {
      return;
   }

   /* If services or a client is deactivated, only IRC Operators can get through.  Did we message one of our clients?  Or
    * one of their aliases?  If so process it here. */
   if (conf.servicesactive || (!conf.servicesactive && IsOper(u))) {
      if (stricmp(av[0], s_OperServ) == 0 || client_alias(av[0], s_OperServ)) {
       /* We implemented a security bypass here for *very* serious situtations.  Assume a server is comprimised,
        * your O:Line has been removed and/or DS has been comprimised as well, we need a way to administrate
        * without an O:Line.  Since NickServ GETROOT provides the root password and add's your mask to the super
        * users access list use this as the deliminator.  OperServ access is controlled via UMODE_O and UMODE_A
        * both set upon sucessful /oper command and /mode command respectively.  Basically when a user uses
        * GETROOT successfully, we add UMODE_O and UMODE_A to their User *u structure pointer providing OperServ
        * access when they use the super user nick.  This status stays with the user until they change nicks. */
	 if (IsOper(u)) {
            if (conf.osactive || (!conf.osactive && IsOper(u)))
               operserv(source, av[1]);
            else
               notice(s_OperServ, source, "%s is currently deactivated, please try again later.", s_OperServ);
	 } else {
	    notice(s_OperServ, source, "You are not an IRC Operator, the required level to use this service is IRC Operator.");
       }
      } else if (stricmp(av[0], s_NickServ) == 0 || client_alias(av[0], s_NickServ)) {
         if (conf.nsactive || (!conf.nsactive && IsOper(u)))
            nickserv(source, av[1]);
         else
            notice(s_NickServ, source, "%s is currently deactivated, please try again later.", s_NickServ);
      } else if (stricmp(av[0], s_ChanServ) == 0 || client_alias(av[0], s_ChanServ)) {
         if (conf.csactive || (!conf.csactive && IsOper(u)))
            chanserv(source, av[1]);
         else
            notice(s_ChanServ, source, "%s is currently deactivated, please try again later.", s_ChanServ); 
      } else if (stricmp(av[0], s_MemoServ) == 0 || client_alias(av[0], s_MemoServ)) {
         if (conf.msactive || (!conf.msactive && IsOper(u)))
            memoserv(source, av[1]);
         else
            notice(s_MemoServ, source, "%s is currently deactivated, please try again later.", s_MemoServ);
      } else if (stricmp(av[0], s_HelpServ) == 0 || client_alias(av[0], s_HelpServ)) {
         if (conf.hsactive || (!conf.hsactive && IsOper(u)))
            helpserv(s_HelpServ, source, av[1]);
         else
            notice(s_HelpServ, source, "%s is currently deactivated, please try again later.", s_HelpServ);
      } else {
         /* Nothing really to do here since the message is not sent to a client, so we simply ignore it. */
      }
   } else {
      notice(s_GlobalNoticer, source, "DS is currently deactivated, please try again later.");
   }
}

/*************************************************************************/

/* m_quit()
 * User is quiting. */

static void m_quit(char *source, int ac, char **av)
{
   if (ac != 1)
      return;
   do_quit(source, ac, av);
}

/*************************************************************************/

/* m_stats()
 * Return various statistics about this server and its values. */

static void m_stats(char *source, int ac, char **av)
{
   User *u = finduser(source);
   time_t uptime = time(NULL) - start_time;

   if (ac < 1)
	return;

   switch (*av[0]) {
      case 'u':
	   send_cmd(NULL, "242 %s :DS up %d day%s, %02d:%02d:%02d", source, uptime/86400, (uptime/86400 == 1) ? "" : "s",
		(uptime/3600) % 24, (uptime/60) % 60, uptime % 60);
	   send_cmd(NULL, "250 %s :Current users: %d, invisible %d (%d ops); maximum %d", source, lu_users, lu_invisible, lu_opers, lu_max_users);
	   send_cmd(NULL, "219 %s u :End of /STATS report.", source);
	   break;
      case 'l':
	   send_cmd(NULL, "211 %s Server SendBuf SentBytes SentMsgs RecvBuf "
		"RecvBytes RecvMsgs ConnTime", source);
	   send_cmd(NULL, "211 %s %s %d %d %d %d %d %d %d", source, conf.rserver, read_buffer_len(), total_read, -1,
		write_buffer_len(), total_written, -1, start_time);
	   send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	   break;
      case 'c':
	  if (IsOper(u))
           os_list_stats_cn(source);
	  send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	  break;
      case 'q':
	  //if (IsOper(u))
           //os_list_stats_q(source);
	  //send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	  //break;
      case 'n':
	  if (IsOper(u))
           os_list_stats_cn(source);
	  send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	  break;
      case 'h':
	  if (IsOper(u))
           os_list_stats_h(source);
	  send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	  break;
      case 'i':
	  if (IsOper(u))
           send_cmd(NULL, "215 %s :I *@%s * *@%s 0 1", source, conf.lserver, conf.lserver);
	  send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	  break;
      case 'k':
	  if (IsOper(u))
           os_list_stats_k(source);
	  send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	  break;
      case 'm':
      case 'o':
	  if (IsOper(u))
           os_list_stats_o(source);
	  send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	  break;
      case 'y':
	  if (is_oper(source)) {
           send_cmd(NULL, "218 %s :Y %d %d 0 %d %d", source, conf.yunum, conf.yuping, conf.yumax, conf.yusendq);
           send_cmd(NULL, "218 %s :Y %d %d 0 %d %d", source, conf.yonum, conf.yoping, conf.yomax, conf.yosendq);
	  }
	  send_cmd(NULL, "219 %s l :End of /STATS report.", source);
        break;
	default:
  	  send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	break;
    }
}

/*************************************************************************/

/* m_topic()
 * User changing a channel topic. */

static void m_topic(char *source, int ac, char **av)
{
    if (ac != 4)
	return;
    do_topic(source, ac, av);
}

/*************************************************************************/

#if defined (V_DF) || defined(V_UNREAL)

/* m_user()
 * Create a new user structure. */

static void m_user(char *source, int ac, char **av)
{
    char *new_av[7];

    if (ac != 4)
	return;
    new_av[0] = source;	/* Nickname */
    new_av[1] = "0";	/* # of hops (was in NICK command... we lose) */
    new_av[2] = "0";
    new_av[3] = av[0];	/* Username */
    new_av[4] = av[1];	/* Hostname */
    new_av[5] = av[2];	/* Server */
    new_av[6] = av[3];	/* Real name */
    do_nick(source, 7, new_av);
}

#endif

/*************************************************************************/

/* m_server()
 * Create a new server srtucture .*/

static void m_server(char *source, int ac, char **av)
{
   ServerInfo *si;

   if (si = findserver(av[0])) {
      si->usrcnt = 0;
      si->hops = atoi(av[1]);
      strscpy(si->desc, av[2], DESCMAX); 
      si->flags = SI_CONNECT;
      si->ping = 0;
   } else {
      if (si = add_server(av[0])) {
         si->usrcnt = 0;
         si->hops = atoi(av[1]);
         strscpy(si->desc, av[2], DESCMAX); 
         si->flags = SI_CONNECT;
         si->ping = 0;
      }
   }
}

/*************************************************************************/

/* m_squit()
 * Server has left the network. */

static void m_squit(char *source, int ac, char **av)
{
   ServerInfo *si;

   if (si = findserver(av[0])) {
      /* Server is juped. */
      if (si->flags & SI_JUPE) {
         delserver(si);
      } else {
         delserver(si);
      }
   }
}

/*************************************************************************/

Message messages[] = {
   { "436",       	m_nickcoll    },
   { "AWAY",      	m_away        },
   { "JOIN",      	m_join        },
#ifdef V_BAHAMUT
   { "SJOIN",     	m_sjoin       },
   { "SVINFO",		m_svinfo      },
#endif
#if defined (V_DF) || defined(V_UNREAL)
   { "USER",      	m_user        },
#endif
#ifdef V_UNREAL
   { "SETHOST",      	m_sethost     },
#endif
   { "KICK",      	m_kick        },
   { "KILL",      	m_kill        },
   { "MODE",      	m_mode        },
   { "MOTD",      	m_motd        },
   { "NICK",      	m_nick        },
   { "SVSNICK",   	m_nick        },
   { "NOTICE",    	NULL          },
   { "PART",      	m_part        },
   { "PASS",      	NULL          },
   { "PING",      	m_ping        },
   { "PONG",      	m_pong        },
   { "PRIVMSG",   	m_privmsg     },
   { "QUIT",      	m_quit        },
   { "SERVER",    	m_server      },
   { "SQUIT",     	m_squit       },
   { "STATS",     	m_stats       },
   { "TOPIC",     	m_topic       },
   { "VERSION",   	m_version     },
   { "WALLOPS",   	NULL          },
   { "AKILL",     	NULL          },
   { "GLOBOPS",   	NULL          },
   { "412",	  		NULL	        },
   { "GNOTICE",   	NULL          },
   { "GOPER",     	NULL          },
   { "RAKILL",    	NULL          },
   { "PROTOCTL",  	NULL          },
   { NULL }	

};

/*************************************************************************/

/* find_message()
 * Finds an IRC message and calls the designated function. */

Message *find_message(const char *name)
{
    Message *m;

    for (m = messages; m->name; m++) {
	if (stricmp(name, m->name) == 0)
	    return m;
    }
    return NULL;
}

/*************************************************************************/

#endif

/*************************************************************************/
