/* Main processing code for Services.
 * src/process.c
 * Copyright (C) 2000-2001 DarkSiderz Net.  All rights reserved.
 * Read the file doc/contact.html for contact information related to
 * this software. */

/*************************************************************************/

#ifndef PROCESS_C
#define PROCESS_C

/*************************************************************************/

#include "services.h"
#include "messages.h"
#include "version.h"

/*************************************************************************/

/* m_version()
 * Response to the VERSION command. */

void m_version(char *source, int ac, char **av)
{
    if (source)
	send_cmd(conf.lserver, "351 %s %s%s %s :%s%s%s%s%s%s%s->%s%s", 
           source, __ABRA_NAME__, __ABRA_VERSION__, conf.lserver,
           (run.flags & RUN_READONLY) == 1 ? "R" : "r",
           conf.servicesactive == 1 ? "S" : "s",
           conf.csactive == 1 ? "C" : "c",
           conf.nsactive == 1 ? "N" : "n",
           conf.msactive == 1 ? "M" : "m",
           conf.osactive == 1 ? "O" : "o",
           conf.hsactive == 1 ? "H" : "h",
       	conf.dumpchannelactive == 1 ? "D" : "d",
       	conf.spcchannelactive == 1 ? "P" : "p");
}

/*************************************************************************/

/* split_buf()
 * Split a buffer into arguments and store the arguments in an
 * argument vector pointed to by argv (which will be malloc'd
 * as necessary); return the argument count.  If colon_special
 * is non-zero, then treat a parameter with a leading ':' as
 * the last parameter of the line, per the IRC RFC.  Destroys
 * the buffer by side effect. */

int split_buf(char *buf, char ***argv, int colon_special)
{
    int argvsize = 8;
    int argc;
    char *s;

    *argv = smalloc(sizeof(char *) * argvsize);
    argc = 0;
    while (*buf) {
	if (argc == argvsize) {
	    argvsize += 8;
	    *argv = srealloc(*argv, sizeof(char *) * argvsize);
	}
	if (*buf == ':') {
	    (*argv)[argc++] = buf+1;
	    buf = "";
	} else {
	    s = strpbrk(buf, " ");
	    if (s) {
		*s++ = 0;
		while (isspace(*s))
		    s++;
	    } else {
		s = buf + strlen(buf);
	    }
	    (*argv)[argc++] = buf;
	    buf = s;
	}
    }
    return argc;
}

/*************************************************************************/

/* process()
 * Main processing routine.  Takes the string in inbuf (global
 * variable) and does something appropriate with it. */

void process()
{
    char source[64];
    char cmd[64];
    char buf[512];		/* Longest legal IRC command line. */
    int ac;				/* Parameters for the command. */
    char **av;
    char *s;
    Message *m;

    /* If debugging, log the buffer. */
   if (debug >= 3)
	log("net: server returned buffer ==> %s", inbuf);

    /* First make a copy of the buffer so we have the original in case we
     * crash - in that case, we want to know what we crashed on. */
    strscpy(buf, inbuf, sizeof(buf));

    /* Split the buffer into pieces. */
    if (*buf == ':') {
	s = strpbrk(buf, " ");
	if (!s)
	    return;
	*s = 0;
	while (isspace(*++s))
	    ;
	strscpy(source, buf+1, sizeof(source));
	strcpy(buf, s);
    } else {
	*source = 0;
    }
    if (!*buf)
	return;
    s = strpbrk(buf, " ");
    if (s) {
	*s = 0;
	while (isspace(*++s))
	    ;
    } else
	s = buf + strlen(buf);
    strscpy(cmd, buf, sizeof(cmd));
    ac = split_buf(s, &av, 1);

    /* Do something with the message, if possible. */
    if (m = find_message(cmd)) {
	if (m->func) {
 	    m->func(source, ac, av);
      }
    } else {
	/* Do not waste space; set to debug mode. */
	if (debug >= 2)
		log("net: server returned unknown message ==> %s", inbuf);
    }

    /* Free argument list we created */
    free(av);
}

/*************************************************************************/

#endif

/*************************************************************************/
