/* Routines for time-delayed actions.
 * src/timeout.c
 * Copyright (C) 2000-2001 Lucas Balemba.  All rights reserved.
 * Read the file doc/contact for contact information related to this software. */

/*************************************************************************/

#ifndef    __timeout_source__
#define __timeout_source__

#ifndef lint
static char sccsid[] = "@(#)timeout.c	7.0 18/03/01 (C) 2001 Lucas Balemba $exp 0";
#endif

/*************************************************************************/

#include "services.h"
#include "timeout.h"

/*************************************************************************/

static Timeout *timeouts = NULL;

/*************************************************************************/

/* check_timeouts()
 * Check the timeout list for any pending actions. */

void check_timeouts(void)
{
   Timeout *to, *to2;
   time_t t = time(NULL);
   User *u;
   NickInfo *ni;

   to = timeouts;
   while (to) {
      if (t < to->timeout) {
	   to = to->next;
	   continue;
	}
	to->code(to);
	if (to->repeat) {
	   to = to->next;
	   continue;
	}
	to2 = to->next;
	if (to->next) {
	   to->next->prev = to->prev;
	}
      if (to->prev) {
         to->prev->next = to->next;
	} else {
	   timeouts = to->next;
      }
	free(to);
	to = to2;
   }
}

/*************************************************************************/

/* add_timeout()
 * Add a timeout to the list to be triggered in `delay' seconds.  If
 * `repeat' is nonzero, do not delete the timeout after it is triggered.
 * This must maintain the property that timeouts added from within a
 * timeout routine do not get checked during that run of the timeout list. */

Timeout *add_timeout(int delay, void (*code)(Timeout *), int repeat)
{
   Timeout *t = smalloc(sizeof(Timeout));
   t->settime = time(NULL);
   t->timeout = t->settime + delay;
   t->code = code;
   t->repeat = repeat;
   t->next = timeouts;
   t->prev = NULL;
   if (timeouts) {
	timeouts->prev = t;
   }
   timeouts = t;
   return t;
}

/*************************************************************************/

/* del_timeout()
 * Remove a timeout from the list (if it's there). */

void del_timeout(Timeout *t)
{
   Timeout *ptr;

   for (ptr = timeouts; ptr; ptr = ptr->next) {
      if (ptr == t)
	   break;
   }
   if (!ptr) {
	return;
   }
   if (t->prev) {
	t->prev->next = t->next;
   } else {
	timeouts = t->next;
   }
   if (t->next) {
	t->next->prev = t->prev;
   }
   free(t);
}

/*************************************************************************/

#endif /* __timeout_source__ */

/*************************************************************************/
